from __future__ import annotations

import dataclasses
import functools
import logging
import time
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar
from uuid import UUID

import psycopg
import sqlalchemy as sql

from pixeltable import exceptions as excs
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator
from pixeltable.metadata import schema

from .dir import Dir
from .globals import IfExistsParam, IfNotExistsParam, MediaValidation
from .insertable_table import InsertableTable
from .path import Path
from .schema_object import SchemaObject
from .table import Table
from .table_version import TableVersion
from .table_version_handle import TableVersionHandle
from .table_version_path import TableVersionPath
from .view import View

if TYPE_CHECKING:
    from .. import DataFrame, exprs


_logger = logging.getLogger('pixeltable')


def _lock_str(for_update: bool) -> str:
    return 'X' if for_update else 'S'


# TODO: remove once the concurrent update behavior has been debugged
# def _debug_print(for_update: bool, msg: str) -> None:
#     return
#     print(f'{datetime.datetime.now()}: {_lock_str(for_update)}: {msg}')


def _unpack_row(
    row: Optional[sql.engine.Row], entities: list[type[sql.orm.decl_api.DeclarativeBase]]
) -> Optional[list[Any]]:
    """Convert a Row result into a list of entity instances.

    Assumes that the query contains a select() of exactly those entities.
    """
    if row is None:
        return None

    result: list[sql.orm.decl_api.DeclarativeBase] = []
    column_offset = 0

    for entity in entities:
        num_cols = len(entity.__table__.columns)
        data = {name: row[column_offset + i] for i, name in enumerate(entity.__table__.columns.keys())}
        inst = entity(**data)
        result.append(inst)
        column_offset += num_cols

    return result


_MAX_RETRIES = 3
T = TypeVar('T')


def _retry_loop(op: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(op)
    def loop(*args: Any, **kwargs: Any) -> T:
        num_remaining_retries = _MAX_RETRIES
        while True:
            try:
                # in order for retry to work, we need to make sure that there aren't any prior db updates
                # that are part of an ongoing transaction
                assert not Env.get().in_xact()
                with Env.get().begin_xact():
                    return op(*args, **kwargs)
            except sql.exc.DBAPIError as e:
                if isinstance(e.orig, psycopg.errors.SerializationFailure) and num_remaining_retries > 0:
                    num_remaining_retries -= 1
                    print(f'serialization failure:\n{e}')
                    print('retrying ************************************************************')
                    time.sleep(1)
                else:
                    raise

    return loop


class Catalog:
    """The functional interface to getting access to catalog objects

    All interface functions must be called in the context of a transaction, started with Env.begin().
    """

    _instance: Optional[Catalog] = None

    # key: [id, version]
    # - mutable version of a table: version == None (even though TableVersion.version is set correctly)
    # - snapshot versions: records the version of the snapshot
    _tbl_versions: dict[tuple[UUID, Optional[int]], TableVersion]
    _tbls: dict[UUID, Table]

    @classmethod
    def get(cls) -> Catalog:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def clear(cls) -> None:
        """Remove the instance. Used for testing."""
        cls._instance = None

    def __init__(self) -> None:
        self._tbl_versions = {}
        self._tbls = {}  # don't use a defaultdict here, it doesn't cooperate with the debugger
        self._init_store()

    def get_dir_path(self, dir_id: UUID) -> Path:
        """Return path for directory with given id"""
        conn = Env.get().conn
        names: list[str] = []
        while True:
            q = sql.select(schema.Dir).where(schema.Dir.id == dir_id)
            # _debug_print(for_update=False, msg=f'dir id={dir_id}')
            row = conn.execute(q).one()
            dir = schema.Dir(**row._mapping)
            if dir.md['name'] == '':
                break
            names.insert(0, dir.md['name'])
            dir_id = dir.parent_id
        return Path('.'.join(names), empty_is_valid=True)

    @dataclasses.dataclass
    class DirEntry:
        dir: Optional[schema.Dir]
        dir_entries: dict[str, Catalog.DirEntry]
        table: Optional[schema.Table]

    @_retry_loop
    def get_dir_contents(self, dir_path: Path, recursive: bool = False) -> dict[str, DirEntry]:
        dir = self._get_schema_object(dir_path, expected=Dir, raise_if_not_exists=True)
        return self._get_dir_contents(dir._id, recursive=recursive)

    def _get_dir_contents(self, dir_id: UUID, recursive: bool = False) -> dict[str, DirEntry]:
        """Returns a dict mapping the entry names to DirEntry objects"""
        conn = Env.get().conn
        result: dict[str, Catalog.DirEntry] = {}

        q = sql.select(schema.Dir).where(schema.Dir.parent_id == dir_id)
        # _debug_print(for_update=False, msg=f'dirs parent_id={dir_id}')
        rows = conn.execute(q).all()
        for row in rows:
            dir = schema.Dir(**row._mapping)
            dir_contents: dict[str, Catalog.DirEntry] = {}
            if recursive:
                dir_contents = self._get_dir_contents(dir.id, recursive=True)
            result[dir.md['name']] = self.DirEntry(dir=dir, dir_entries=dir_contents, table=None)

        q = sql.select(schema.Table).where(schema.Table.dir_id == dir_id)
        # _debug_print(for_update=False, msg=f'tbls parent_id={dir_id}')
        rows = conn.execute(q).all()
        for row in rows:
            tbl = schema.Table(**row._mapping)
            result[tbl.md['name']] = self.DirEntry(dir=None, dir_entries={}, table=tbl)

        return result

    @_retry_loop
    def move(self, path: Path, new_path: Path) -> None:
        _, dest_dir, src_obj = self._prepare_dir_op(
            add_dir_path=new_path.parent,
            add_name=new_path.name,
            drop_dir_path=path.parent,
            drop_name=path.name,
            raise_if_exists=True,
            raise_if_not_exists=True,
        )
        src_obj._move(new_path.name, dest_dir._id)

    def _prepare_dir_op(
        self,
        add_dir_path: Optional[Path] = None,
        add_name: Optional[str] = None,
        drop_dir_path: Optional[Path] = None,
        drop_name: Optional[str] = None,
        drop_expected: Optional[type[SchemaObject]] = None,
        raise_if_exists: bool = False,
        raise_if_not_exists: bool = False,
    ) -> tuple[Optional[SchemaObject], Optional[SchemaObject], Optional[SchemaObject]]:
        """
        Validates paths and acquires locks needed for a directory operation, ie, add/drop/rename (add + drop) of a
        directory entry.

        The target entry is either a table or directory. The directory operation can include
        - adding an entry (<add_dir_path>.<add_name>)
        - dropping an entry (<drop_dir_path>.<drop_name>)

        Returns: (existing SchemaObject of add path, Dir of add path, existing SchemaObject of drop path)

        Locking protocol:
        - X locks on the immediate parent directories of the added/dropped entries; this prevents concurrent
          modifications of the parent
        - lock parent before child
        - if both add and drop (= two directories are involved), lock the directories in a pre-determined order
          (in this case, by name) in order to prevent deadlocks between concurrent directory modifications
        """
        assert (add_dir_path is None) == (add_name is None)
        assert (drop_dir_path is None) == (drop_name is None)
        dir_paths: set[Path] = set()
        if add_dir_path is not None:
            dir_paths.add(add_dir_path)
        if drop_dir_path is not None:
            dir_paths.add(drop_dir_path)

        add_dir: Optional[schema.Dir] = None
        drop_dir: Optional[schema.Dir] = None
        for p in sorted(dir_paths):
            dir = self._get_dir(p, for_update=True)
            if dir is None:
                raise excs.Error(f'Directory {str(p)!r} does not exist.')
            if p == add_dir_path:
                add_dir = dir
            if p == drop_dir_path:
                drop_dir = dir

        add_obj: Optional[SchemaObject] = None
        if add_dir is not None:
            add_obj = self._get_dir_entry(add_dir.id, add_name, for_update=True)
            if add_obj is not None and raise_if_exists:
                add_path = add_dir_path.append(add_name)
                raise excs.Error(f'Path {str(add_path)!r} already exists.')

        drop_obj: Optional[SchemaObject] = None
        if drop_dir is not None:
            drop_path = drop_dir_path.append(drop_name)
            drop_obj = self._get_dir_entry(drop_dir.id, drop_name, for_update=True)
            if drop_obj is None and raise_if_not_exists:
                raise excs.Error(f'Path {str(drop_path)!r} does not exist.')
            if drop_obj is not None and drop_expected is not None and not isinstance(drop_obj, drop_expected):
                raise excs.Error(
                    f'{str(drop_path)!r} needs to be a {drop_expected._display_name()} '
                    f'but is a {type(drop_obj)._display_name()}'
                )

        add_dir_obj = Dir(add_dir.id, add_dir.parent_id, add_dir.md['name']) if add_dir is not None else None
        return add_obj, add_dir_obj, drop_obj

    def _get_dir_entry(self, dir_id: UUID, name: str, for_update: bool = False) -> Optional[SchemaObject]:
        user = Env.get().user
        conn = Env.get().conn

        # check for subdirectory
        q = sql.select(schema.Dir).where(
            schema.Dir.parent_id == dir_id, schema.Dir.md['name'].astext == name, schema.Dir.md['user'].astext == user
        )
        if for_update:
            q = q.with_for_update()
        # _debug_print(for_update, f'dir name={name!r} parent={dir_id}')
        # row = conn.execute(q).one_or_none()
        # if row is not None:
        #     dir_record = schema.Dir(**row._mapping)
        #     return Dir(dir_record.id, dir_record.parent_id, name)
        rows = conn.execute(q).all()
        if len(rows) > 1:
            raise AssertionError(rows)
        if len(rows) == 1:
            dir_record = schema.Dir(**rows[0]._mapping)
            return Dir(dir_record.id, dir_record.parent_id, name)

        # check for table
        q = sql.select(schema.Table.id).where(
            schema.Table.dir_id == dir_id,
            schema.Table.md['name'].astext == name,
            schema.Table.md['user'].astext == user,
        )
        if for_update:
            q = q.with_for_update()
        # _debug_print(for_update, f'table name={name!r} parent={dir_id}')
        tbl_id = conn.execute(q).scalar_one_or_none()
        if tbl_id is not None:
            if tbl_id not in self._tbls:
                self._tbls[tbl_id] = self._load_tbl(tbl_id)
            return self._tbls[tbl_id]

        return None

    def _get_schema_object(
        self,
        path: Path,
        expected: Optional[type[SchemaObject]] = None,
        raise_if_exists: bool = False,
        raise_if_not_exists: bool = False,
        for_update: bool = False,
    ) -> Optional[SchemaObject]:
        """Return the schema object at the given path, or None if it doesn't exist.

        Raises Error if
        - the parent directory doesn't exist'
        - raise_if_exists is True and the path exists
        - raise_if_not_exists is True and the path does not exist
        - expected is not None and the existing object has a different type
        """
        if path.is_root:
            # the root dir
            if expected is not None and expected is not Dir:
                raise excs.Error(
                    f'{str(path)!r} needs to be a {expected._display_name()} but is a {Dir._display_name()}'
                )
            dir = self._get_dir(path, for_update=for_update)
            if dir is None:
                raise excs.Error(f'Unknown user: {Env.get().user}')
            return Dir(dir.id, dir.parent_id, dir.md['name'])

        parent_path = path.parent
        parent_dir = self._get_dir(parent_path, for_update=False)
        if parent_dir is None:
            raise excs.Error(f'Directory {str(parent_path)!r} does not exist.')
        obj = self._get_dir_entry(parent_dir.id, path.name, for_update=for_update)

        if obj is None and raise_if_not_exists:
            raise excs.Error(f'Path {str(path)!r} does not exist.')
        elif obj is not None and raise_if_exists:
            raise excs.Error(f'Path {str(path)!r} is an existing {type(obj)._display_name()}.')
        elif obj is not None and expected is not None and not isinstance(obj, expected):
            raise excs.Error(
                f'{str(path)!r} needs to be a {expected._display_name()} but is a {type(obj)._display_name()}.'
            )
        return obj

    def get_table_by_id(self, tbl_id: UUID) -> Optional[Table]:
        if tbl_id not in self._tbls:
            tbl = self._load_tbl(tbl_id)
            if tbl is None:
                return None
            self._tbls[tbl_id] = tbl
        return self._tbls[tbl_id]

    @_retry_loop
    def create_table(
        self,
        path: Path,
        schema: dict[str, Any],
        df: 'DataFrame',
        if_exists: IfExistsParam,
        primary_key: Optional[list[str]],
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
    ) -> Table:
        existing = self._handle_path_collision(path, InsertableTable, False, if_exists)
        if existing is not None:
            assert isinstance(existing, Table)
            return existing

        dir = self._get_schema_object(path.parent, expected=Dir, raise_if_not_exists=True)
        assert dir is not None

        tbl = InsertableTable._create(
            dir._id,
            path.name,
            schema,
            df,
            primary_key=primary_key,
            num_retained_versions=num_retained_versions,
            comment=comment,
            media_validation=media_validation,
        )
        self._tbls[tbl._id] = tbl
        return tbl

    @_retry_loop
    def create_view(
        self,
        path: Path,
        base: TableVersionPath,
        select_list: Optional[list[tuple[exprs.Expr, Optional[str]]]],
        where: Optional[exprs.Expr],
        additional_columns: Optional[dict[str, Any]],
        is_snapshot: bool,
        iterator: Optional[tuple[type[ComponentIterator], dict[str, Any]]],
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
        if_exists: IfExistsParam,
    ) -> Table:
        from pixeltable.utils.filecache import FileCache

        existing = self._handle_path_collision(path, View, is_snapshot, if_exists)
        if existing is not None:
            assert isinstance(existing, View)
            return existing

        dir = self._get_schema_object(path.parent, expected=Dir, raise_if_not_exists=True)
        assert dir is not None
        if iterator is None:
            iterator_class, iterator_args = None, None
        else:
            iterator_class, iterator_args = iterator
        view = View._create(
            dir._id,
            path.name,
            base=base,
            select_list=select_list,
            additional_columns=additional_columns,
            predicate=where,
            is_snapshot=is_snapshot,
            iterator_cls=iterator_class,
            iterator_args=iterator_args,
            num_retained_versions=num_retained_versions,
            comment=comment,
            media_validation=media_validation,
        )
        FileCache.get().emit_eviction_warnings()
        self._tbls[view._id] = view
        return view

    @_retry_loop
    def get_table(self, path: Path) -> Table:
        obj = Catalog.get()._get_schema_object(path, expected=Table, raise_if_not_exists=True)
        assert isinstance(obj, Table)
        obj._tbl_version.get().ensure_md_loaded()
        return obj

    @_retry_loop
    def drop_table(self, path: Path, if_not_exists: IfNotExistsParam, force: bool) -> None:
        _, _, src_obj = self._prepare_dir_op(
            drop_dir_path=path.parent,
            drop_name=path.name,
            drop_expected=Table,
            raise_if_not_exists=if_not_exists == IfNotExistsParam.ERROR and not force,
        )
        if src_obj is None:
            _logger.info(f'Skipped table {str(path)!r} (does not exist).')
            return
        assert isinstance(src_obj, Table)
        self._drop_tbl(src_obj, force=force, is_replace=False)

    def _drop_tbl(self, tbl: Table, force: bool, is_replace: bool) -> None:
        """
        Drop the table (and recursively its views, if force == True).

        Locking protocol:
        - X-lock base before X-locking any view
        - deadlock-free wrt to TableVersion.insert() (insert propagation also proceeds top-down)
        - X-locks parent dir prior to calling TableVersion.drop(): prevent concurrent creation of another SchemaObject
          in the same directory with the same name (which could lead to duplicate names if we get rolled back)
        """
        view_ids = self.get_view_ids(tbl._id, for_update=True)
        if len(view_ids) > 0:
            if not force:
                is_snapshot = tbl._tbl_version_path.is_snapshot()
                obj_type_str = 'Snapshot' if is_snapshot else tbl._display_name().capitalize()
                msg: str
                if is_replace:
                    msg = (
                        f'{obj_type_str} {tbl._path()} already exists and has dependents. '
                        "Use `if_exists='replace_force'` to replace it."
                    )
                else:
                    msg = f'{obj_type_str} {tbl._path()} has dependents.'
                raise excs.Error(msg)

            for view_id in view_ids:
                view = self.get_table_by_id(view_id)
                self._drop_tbl(view, force=force, is_replace=is_replace)

        _ = self.get_dir(tbl._dir_id, for_update=True)  # X-lock the parent directory
        tbl._drop()
        assert tbl._id in self._tbls
        del self._tbls[tbl._id]
        _logger.info(f'Dropped table `{tbl._path()}`.')

    @_retry_loop
    def create_dir(self, path: Path, if_exists: IfExistsParam, parents: bool) -> Dir:
        # existing = self._handle_path_collision(path, Dir, False, if_exists)
        # if existing is not None:
        #     assert isinstance(existing, Dir)
        #     return existing
        #
        # parent = self._get_schema_object(path.parent)
        # assert parent is not None
        # dir = Dir._create(parent._id, path.name)
        # Env.get().console_logger.info(f'Created directory {str(path)!r}.')
        # return dir

        if parents:
            # start walking down from the root
            last_parent: Optional[SchemaObject] = None
            for ancestor in path.ancestors():
                ancestor_obj = self._get_schema_object(ancestor, expected=Dir)
                assert ancestor_obj is not None or last_parent is not None
                last_parent = Dir._create(last_parent._id, ancestor.name) if ancestor_obj is None else ancestor_obj
            parent = last_parent
        else:
            parent = self._get_schema_object(path.parent)
        existing = self._handle_path_collision(path, Dir, False, if_exists)
        if existing is not None:
            assert isinstance(existing, Dir)
            return existing
        assert parent is not None
        dir = Dir._create(parent._id, path.name)
        Env.get().console_logger.info(f'Created directory {str(path)!r}.')
        return dir

    @_retry_loop
    def drop_dir(self, path: Path, if_not_exists: IfNotExistsParam, force: bool) -> None:
        _, _, schema_obj = self._prepare_dir_op(
            drop_dir_path=path.parent,
            drop_name=path.name,
            drop_expected=Dir,
            raise_if_not_exists=if_not_exists == IfNotExistsParam.ERROR and not force,
        )
        if schema_obj is None:
            _logger.info(f'Directory {str(path)!r} does not exist; skipped drop_dir().')
            return
        self._drop_dir(schema_obj._id, path, force=force)

    def _drop_dir(self, dir_id: UUID, dir_path: Path, force: bool = False) -> None:
        conn = Env.get().conn
        if not force:
            # check for existing entries
            q = sql.select(sql.func.count()).select_from(schema.Dir).where(schema.Dir.parent_id == dir_id)
            num_subdirs = conn.execute(q).scalar()
            q = sql.select(sql.func.count()).select_from(schema.Table).where(schema.Table.dir_id == dir_id)
            num_tbls = conn.execute(q).scalar()
            if num_subdirs + num_tbls > 0:
                raise excs.Error(f'Directory {str(dir_path)!r} is not empty.')

        # drop existing subdirs
        dir_q = sql.select(schema.Dir).where(schema.Dir.parent_id == dir_id).with_for_update()
        for row in conn.execute(dir_q).all():
            self._drop_dir(row.id, dir_path.append(row.md['name']), force=True)

        # drop existing tables
        tbl_q = sql.select(schema.Table).where(schema.Table.dir_id == dir_id).with_for_update()
        for row in conn.execute(tbl_q).all():
            tbl = self.get_table_by_id(row.id)
            # this table would have been dropped already if it's a view of a base we dropped earlier
            if tbl is not None:
                self._drop_tbl(tbl, force=True, is_replace=False)

        # self.drop_dir(dir_id)
        # _debug_print(for_update=True, msg=f'drop dir id={dir_id}')
        conn.execute(sql.delete(schema.Dir).where(schema.Dir.id == dir_id))
        _logger.info(f'Removed directory {str(dir_path)!r}.')

    def get_view_ids(self, tbl_id: UUID, for_update: bool = False) -> list[UUID]:
        """Return the ids of views that directly reference the given table"""
        conn = Env.get().conn
        q = sql.select(schema.Table.id).where(sql.text(f"md->'view_md'->'base_versions'->0->>0 = {tbl_id.hex!r}"))
        if for_update:
            q = q.with_for_update()
        # _debug_print(for_update=False, msg=f'views of tbl id={tbl_id}')
        result = [r[0] for r in conn.execute(q).all()]
        return result

    def get_tbl_version(self, tbl_id: UUID, effective_version: Optional[int]) -> Optional[TableVersion]:
        if (tbl_id, effective_version) not in self._tbl_versions:
            self._tbl_versions[tbl_id, effective_version] = self._load_tbl_version(tbl_id, effective_version)
        return self._tbl_versions[tbl_id, effective_version]

    def add_tbl_version(self, tbl_version: TableVersion) -> None:
        """Explicitly add a TableVersion"""
        self._tbl_versions[tbl_version.id, tbl_version.effective_version] = tbl_version
        # if this is a mutable view, also record it in the base
        if tbl_version.is_view and tbl_version.effective_version is None:
            base = tbl_version.base.get()
            base.mutable_views.append(TableVersionHandle(tbl_version.id, tbl_version.effective_version))

    def remove_tbl_version(self, tbl_version: TableVersion) -> None:
        assert (tbl_version.id, tbl_version.effective_version) in self._tbl_versions
        del self._tbl_versions[tbl_version.id, tbl_version.effective_version]

    def get_dir(self, dir_id: UUID, for_update: bool = False) -> Optional[Dir]:
        """Return the Dir with the given id, or None if it doesn't exist"""
        conn = Env.get().conn
        q = sql.select(schema.Dir).where(schema.Dir.id == dir_id)
        if for_update:
            q = q.with_for_update()
        # _debug_print(for_update=False, msg=f'dir id={dir_id!r}')
        row = conn.execute(q).one_or_none()
        if row is None:
            return None
        dir_record = schema.Dir(**row._mapping)
        return Dir(dir_record.id, dir_record.parent_id, dir_record.md['name'])

    def _get_dir(self, path: Path, for_update: bool = False) -> Optional[schema.Dir]:
        """
        Locking protocol:
        - S locks on all ancestors
        - X lock on dir if for_update == True, otherwise also an S lock
        """
        user = Env.get().user
        conn = Env.get().conn
        if path.is_root:
            q = sql.select(schema.Dir).where(schema.Dir.parent_id.is_(None), schema.Dir.md['user'].astext == user)
            if for_update:
                q = q.with_for_update()
            # _debug_print(for_update, 'root dir')
            row = conn.execute(q).one_or_none()
            return schema.Dir(**row._mapping) if row is not None else None
        else:
            parent_dir = self._get_dir(path.parent, for_update=False)
            if parent_dir is None:
                return None
            q = sql.select(schema.Dir).where(
                schema.Dir.parent_id == parent_dir.id,
                schema.Dir.md['name'].astext == path.name,
                schema.Dir.md['user'].astext == user,
            )
            if for_update:
                q = q.with_for_update()
            # _debug_print(for_update, f'dir {str(path)}')
            row = conn.execute(q).one_or_none()
            return schema.Dir(**row._mapping) if row is not None else None

    def _load_tbl(self, tbl_id: UUID) -> Optional[Table]:
        _logger.info(f'Loading table {tbl_id}')
        from .insertable_table import InsertableTable
        from .view import View

        conn = Env.get().conn
        q = (
            sql.select(schema.Table, schema.TableSchemaVersion)
            .join(schema.TableSchemaVersion)
            .where(schema.Table.id == schema.TableSchemaVersion.tbl_id)
            # Table.md['current_schema_version'] == TableSchemaVersion.schema_version
            .where(
                sql.text(
                    f"({schema.Table.__table__}.md->>'current_schema_version')::int = "
                    f'{schema.TableSchemaVersion.__table__}.{schema.TableSchemaVersion.schema_version.name}'
                )
            )
            .where(schema.Table.id == tbl_id)
        )
        # _debug_print(for_update=False, msg=f'load table id={tbl_id!r}')
        row = conn.execute(q).one_or_none()
        if row is None:
            return None
        tbl_record, schema_version_record = _unpack_row(row, [schema.Table, schema.TableSchemaVersion])

        tbl_md = schema.md_from_dict(schema.TableMd, tbl_record.md)
        view_md = tbl_md.view_md
        if view_md is None:
            # this is a base table
            if (tbl_id, None) not in self._tbl_versions:
                self._tbl_versions[tbl_id, None] = self._load_tbl_version(tbl_id, None)
            tbl = InsertableTable(tbl_record.dir_id, TableVersionHandle(tbl_id, None))
            return tbl

        # this is a view; determine the sequence of TableVersions to load
        tbl_version_path: list[tuple[UUID, Optional[int]]] = []
        schema_version_md = schema.md_from_dict(schema.TableSchemaVersionMd, schema_version_record.md)
        pure_snapshot = view_md.is_snapshot and view_md.predicate is None and len(schema_version_md.columns) == 0
        if pure_snapshot:
            # this is a pure snapshot, without a physical table backing it; we only need the bases
            pass
        else:
            effective_version = 0 if view_md.is_snapshot else None  # snapshots only have version 0
            tbl_version_path.append((tbl_id, effective_version))
        tbl_version_path.extend((UUID(tbl_id), version) for tbl_id, version in view_md.base_versions)

        # load TableVersions, starting at the root
        base_path: Optional[TableVersionPath] = None
        view_path: Optional[TableVersionPath] = None
        for id, effective_version in tbl_version_path[::-1]:
            if (id, effective_version) not in self._tbl_versions:
                self._tbl_versions[id, effective_version] = self._load_tbl_version(id, effective_version)
            view_path = TableVersionPath(TableVersionHandle(id, effective_version), base=base_path)
            base_path = view_path
        view = View(tbl_id, tbl_record.dir_id, tbl_md.name, view_path, snapshot_only=pure_snapshot)
        # TODO: also load mutable views
        return view

    def _load_tbl_version(self, tbl_id: UUID, effective_version: Optional[int]) -> Optional[TableVersion]:
        _logger.info(f'Loading table version: {tbl_id}:{effective_version}')
        conn = Env.get().conn
        q = (
            sql.select(schema.Table, schema.TableSchemaVersion)
            .select_from(schema.Table)
            .where(schema.Table.id == tbl_id)
            .join(schema.TableSchemaVersion)
            .where(schema.TableSchemaVersion.tbl_id == tbl_id)
        )

        if effective_version is not None:
            # we are loading a specific version
            # SELECT *
            # FROM Table t
            # JOIN TableVersion tv ON (tv.tbl_id = tbl_id AND tv.version = effective_version)
            # JOIN TableSchemaVersion tsv ON (tsv.tbl_id = tbl_id AND tv.md.schema_version = tsv.schema_version)
            # WHERE t.id = tbl_id
            q = (
                q.join(schema.TableVersion)
                .where(schema.TableVersion.tbl_id == tbl_id)
                .where(sql.text(f"({schema.TableVersion.__table__}.md->>'version')::int = {effective_version}"))
                .where(
                    sql.text(
                        (
                            f"({schema.TableVersion.__table__}.md->>'schema_version')::int = "
                            f'{schema.TableSchemaVersion.__table__}.{schema.TableSchemaVersion.schema_version.name}'
                        )
                    )
                )
            )
        else:
            # we are loading the current version
            # SELECT *
            # FROM Table t
            # JOIN TableSchemaVersion tsv ON (tsv.tbl_id = tbl_id AND t.current_schema_version = tsv.schema_version)
            # WHERE t.id = tbl_id
            q = q.where(
                sql.text(
                    (
                        f"({schema.Table.__table__}.md->>'current_schema_version')::int = "
                        f'{schema.TableSchemaVersion.__table__}.{schema.TableSchemaVersion.schema_version.name}'
                    )
                )
            )

        row = conn.execute(q).one_or_none()
        tbl_record, schema_version_record = _unpack_row(row, [schema.Table, schema.TableSchemaVersion])
        tbl_md = schema.md_from_dict(schema.TableMd, tbl_record.md)
        schema_version_md = schema.md_from_dict(schema.TableSchemaVersionMd, schema_version_record.md)
        view_md = tbl_md.view_md

        # load mutable view ids
        q = sql.select(schema.Table.id).where(
            sql.text(
                f"md->'view_md'->'base_versions'->0->>0 = {tbl_id.hex!r} "
                "AND md->'view_md'->'base_versions'->0->1 IS NULL"
            )
        )
        mutable_view_ids = [r[0] for r in conn.execute(q).all()]
        mutable_views = [TableVersionHandle(id, None) for id in mutable_view_ids]

        if view_md is None:
            # this is a base table
            tbl_version = TableVersion(
                tbl_record.id, tbl_md, effective_version, schema_version_md, mutable_views=mutable_views
            )
            return tbl_version

        assert len(view_md.base_versions) > 0  # a view needs to have a base
        pure_snapshot = view_md.is_snapshot and view_md.predicate is None and len(schema_version_md.columns) == 0
        assert not pure_snapshot  # a pure snapshot doesn't have a physical table backing it, no point in loading it

        base: TableVersionHandle
        base_path: Optional[TableVersionPath] = None  # needed for live view
        if view_md.is_snapshot:
            base = TableVersionHandle(UUID(view_md.base_versions[0][0]), view_md.base_versions[0][1])
        else:
            base_path = TableVersionPath.from_md(tbl_md.view_md.base_versions)
            base = base_path.tbl_version

        tbl_version = TableVersion(
            tbl_record.id,
            tbl_md,
            effective_version,
            schema_version_md,
            base_path=base_path,
            base=base,
            mutable_views=mutable_views,
        )
        return tbl_version

    def _init_store(self) -> None:
        """One-time initialization of the stored catalog. Idempotent."""
        self.create_user(None)
        _logger.info('Initialized catalog.')

    def create_user(self, user: Optional[str]) -> None:
        """
        Creates a catalog record (root directory) for the specified user, if one does not already exist.
        """
        with Env.get().begin_xact():
            session = Env.get().session
            # See if there are any directories in the catalog matching the specified user.
            if session.query(schema.Dir).where(schema.Dir.md['user'].astext == user).count() > 0:
                # At least one such directory exists; no need to create a new one.
                return

            dir_md = schema.DirMd(name='', user=user, additional_md={})
            dir_record = schema.Dir(parent_id=None, md=dataclasses.asdict(dir_md))
            session.add(dir_record)
            session.flush()
            _logger.info(f'Added root directory record for user: {user!r}')

    def _handle_path_collision(
        self, path: Path, expected_obj_type: type[SchemaObject], expected_snapshot: bool, if_exists: IfExistsParam
    ) -> Optional[SchemaObject]:
        obj, _, _ = self._prepare_dir_op(add_dir_path=path.parent, add_name=path.name)

        if if_exists == IfExistsParam.ERROR and obj is not None:
            raise excs.Error(f'Path {str(path)!r} is an existing {type(obj)._display_name()}')
        else:
            is_snapshot = isinstance(obj, View) and obj._tbl_version_path.is_snapshot()
            if obj is not None and (not isinstance(obj, expected_obj_type) or (expected_snapshot and not is_snapshot)):
                obj_type_str = 'snapshot' if expected_snapshot else expected_obj_type._display_name()
                raise excs.Error(
                    f'Path {str(path)!r} already exists but is not a {obj_type_str}. '
                    f'Cannot {if_exists.name.lower()} it.'
                )

        if obj is None:
            return None
        if if_exists == IfExistsParam.IGNORE:
            return obj

        # drop the existing schema object
        if isinstance(obj, Dir):
            dir_contents = self._get_dir_contents(obj._id)
            if len(dir_contents) > 0 and if_exists == IfExistsParam.REPLACE:
                raise excs.Error(
                    f'Directory {str(path)!r} already exists and is not empty. '
                    'Use `if_exists="replace_force"` to replace it.'
                )
            self._drop_dir(obj._id, path, force=True)
        else:
            assert isinstance(obj, Table)
            self._drop_tbl(obj, force=if_exists == IfExistsParam.REPLACE_FORCE, is_replace=True)
        return None



import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.schema import Dir, Table, TableSchemaVersion, TableVersion


@register_converter(version=28)
def _(engine: sql.engine.Engine) -> None:
    with engine.begin() as conn:
        conn.execute(sql.update(Dir).values(md=Dir.md.concat({'user': None, 'additional_md': {}})))
        conn.execute(sql.update(Table).values(md=Table.md.concat({'user': None, 'additional_md': {}})))
        conn.execute(sql.update(TableVersion).values(md=TableVersion.md.concat({'additional_md': {}})))
        conn.execute(sql.update(TableSchemaVersion).values(md=TableSchemaVersion.md.concat({'additional_md': {}})))



from __future__ import annotations

import datetime
import glob
import http.server
import importlib
import importlib.util
import inspect
import logging
import os
import platform
import shutil
import subprocess
import sys
import threading
import uuid
import warnings
from abc import abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from sys import stdout
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, TypeVar
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import pixeltable_pgserver
import sqlalchemy as sql
from tqdm import TqdmWarning

from pixeltable import exceptions as excs
from pixeltable.config import Config
from pixeltable.utils.console_output import ConsoleLogger, ConsoleMessageFilter, ConsoleOutputHandler, map_level
from pixeltable.utils.dbms import CockroachDbms, Dbms, PostgresqlDbms
from pixeltable.utils.http_server import make_server

if TYPE_CHECKING:
    import spacy


_logger = logging.getLogger('pixeltable')

T = TypeVar('T')


class Env:
    """
    Store runtime globals for both local and non-local environments.
    For a local environment, Pixeltable uses an embedded PostgreSQL server that runs locally in a separate process.
    For a non-local environment, Pixeltable uses a connection string to the externally managed database.
    """

    _instance: Optional[Env] = None
    __initializing: bool = False
    _log_fmt_str = '%(asctime)s %(levelname)s %(name)s %(filename)s:%(lineno)d: %(message)s'

    _media_dir: Optional[Path]
    _file_cache_dir: Optional[Path]  # cached media files with external URL
    _dataset_cache_dir: Optional[Path]  # cached datasets (eg, pytorch or COCO)
    _log_dir: Optional[Path]  # log files
    _tmp_dir: Optional[Path]  # any tmp files
    _sa_engine: Optional[sql.engine.base.Engine]
    _pgdata_dir: Optional[Path]
    _db_name: Optional[str]
    _db_server: Optional[pixeltable_pgserver.PostgresServer]  # set only when running in local environment
    _db_url: Optional[str]
    _default_time_zone: Optional[ZoneInfo]

    # info about optional packages that are utilized by some parts of the code
    __optional_packages: dict[str, PackageInfo]

    _spacy_nlp: Optional[spacy.Language]
    _httpd: Optional[http.server.HTTPServer]
    _http_address: Optional[str]
    _logger: logging.Logger
    _default_log_level: int
    _logfilename: Optional[str]
    _log_to_stdout: bool
    _module_log_level: dict[str, int]  # module name -> log level
    _file_cache_size_g: float
    _pxt_api_key: Optional[str]
    _stdout_handler: logging.StreamHandler
    _initialized: bool

    _resource_pool_info: dict[str, Any]
    _current_conn: Optional[sql.Connection]
    _current_session: Optional[sql.orm.Session]
    _dbms: Optional[Dbms]

    @classmethod
    def get(cls) -> Env:
        if cls._instance is None:
            cls._init_env()
        return cls._instance

    @classmethod
    def _init_env(cls, reinit_db: bool = False) -> None:
        assert not cls.__initializing, 'Circular env initialization detected.'
        cls.__initializing = True
        env = Env()
        env._set_up(reinit_db=reinit_db)
        env._upgrade_metadata()
        cls._instance = env
        cls.__initializing = False

    def __init__(self) -> None:
        assert self._instance is None, 'Env is a singleton; use Env.get() to access the instance'

        self._media_dir = None  # computed media files
        self._file_cache_dir = None  # cached media files with external URL
        self._dataset_cache_dir = None  # cached datasets (eg, pytorch or COCO)
        self._log_dir = None  # log files
        self._tmp_dir = None  # any tmp files
        self._sa_engine = None
        self._pgdata_dir = None
        self._db_name = None
        self._db_server = None
        self._db_url = None
        self._default_time_zone = None
        self.__optional_packages = {}
        self._spacy_nlp = None
        self._httpd = None
        self._http_address = None

        # logging-related state
        self._logger = logging.getLogger('pixeltable')
        self._logger.setLevel(logging.DEBUG)  # allow everything to pass, we filter in _log_filter()
        self._logger.propagate = False
        self._logger.addFilter(self._log_filter)
        self._default_log_level = logging.INFO
        self._logfilename = None
        self._log_to_stdout = False
        self._module_log_level = {}  # module name -> log level

        # create logging handler to also log to stdout
        self._stdout_handler = logging.StreamHandler(stream=sys.stdout)
        self._stdout_handler.setFormatter(logging.Formatter(self._log_fmt_str))
        self._initialized = False

        self._resource_pool_info = {}
        self._current_conn = None
        self._current_session = None
        self._dbms = None

    @property
    def db_url(self) -> str:
        assert self._db_url is not None
        return self._db_url

    @property
    def http_address(self) -> str:
        assert self._http_address is not None
        return self._http_address

    @property
    def user(self) -> Optional[str]:
        return Config.get().get_string_value('user')

    @user.setter
    def user(self, user: Optional[str]) -> None:
        if user is None:
            if 'PIXELTABLE_USER' in os.environ:
                del os.environ['PIXELTABLE_USER']
        else:
            os.environ['PIXELTABLE_USER'] = user

    @property
    def default_time_zone(self) -> Optional[ZoneInfo]:
        return self._default_time_zone

    @default_time_zone.setter
    def default_time_zone(self, tz: Optional[ZoneInfo]) -> None:
        """
        This is not a publicly visible setter; it is only for testing purposes.
        """
        tz_name = None if tz is None else tz.key
        self.engine.dispose()
        self._create_engine(time_zone_name=tz_name)

    @property
    def conn(self) -> Optional[sql.Connection]:
        assert self._current_conn is not None
        return self._current_conn

    @property
    def session(self) -> Optional[sql.orm.Session]:
        assert self._current_session is not None
        return self._current_session

    @property
    def dbms(self) -> Optional[Dbms]:
        assert self._dbms is not None
        return self._dbms

    def in_xact(self) -> bool:
        return self._current_conn is not None

    @property
    def is_local(self) -> bool:
        assert self._db_url is not None  # is_local should be called only after db initialization
        return self._db_server is not None

    @contextmanager
    def begin_xact(self) -> Iterator[sql.Connection]:
        """Return a context manager that yields a connection to the database. Idempotent."""
        if self._current_conn is None:
            assert self._current_session is None
            try:
                with self.engine.begin() as conn, sql.orm.Session(conn) as session:
                    # TODO: remove print() once we're done with debugging the concurrent update behavior
                    # print(f'{datetime.datetime.now()}: start xact')
                    self._current_conn = conn
                    self._current_session = session
                    yield conn
            finally:
                self._current_session = None
                self._current_conn = None
                # print(f'{datetime.datetime.now()}: end xact')
        else:
            assert self._current_session is not None
            yield self._current_conn

    def configure_logging(
        self,
        *,
        to_stdout: Optional[bool] = None,
        level: Optional[int] = None,
        add: Optional[str] = None,
        remove: Optional[str] = None,
    ) -> None:
        """Configure logging.

        Args:
            to_stdout: if True, also log to stdout
            level: default log level
            add: comma-separated list of 'module name:log level' pairs; ex.: add='video:10'
            remove: comma-separated list of module names
        """
        if to_stdout is not None:
            self.log_to_stdout(to_stdout)
        if level is not None:
            self.set_log_level(level)
        if add is not None:
            for module, level_str in [t.split(':') for t in add.split(',')]:
                self.set_module_log_level(module, int(level_str))
        if remove is not None:
            for module in remove.split(','):
                self.set_module_log_level(module, None)
        if to_stdout is None and level is None and add is None and remove is None:
            self.print_log_config()

    def print_log_config(self) -> None:
        print(f'logging to {self._logfilename}')
        print(f'{"" if self._log_to_stdout else "not "}logging to stdout')
        print(f'default log level: {logging.getLevelName(self._default_log_level)}')
        print(
            f'module log levels: '
            f'{",".join([name + ":" + logging.getLevelName(val) for name, val in self._module_log_level.items()])}'
        )

    def log_to_stdout(self, enable: bool = True) -> None:
        self._log_to_stdout = enable
        if enable:
            self._logger.addHandler(self._stdout_handler)
        else:
            self._logger.removeHandler(self._stdout_handler)

    def set_log_level(self, level: int) -> None:
        self._default_log_level = level

    def set_module_log_level(self, module: str, level: Optional[int]) -> None:
        if level is None:
            self._module_log_level.pop(module, None)
        else:
            self._module_log_level[module] = level

    def is_installed_package(self, package_name: str) -> bool:
        assert package_name in self.__optional_packages
        return self.__optional_packages[package_name].is_installed

    def _log_filter(self, record: logging.LogRecord) -> bool:
        if record.name == 'pixeltable':
            # accept log messages from a configured pixeltable module (at any level of the module hierarchy)
            path_parts = list(Path(record.pathname).parts)
            path_parts.reverse()
            max_idx = path_parts.index('pixeltable')
            for module_name in path_parts[:max_idx]:
                if module_name in self._module_log_level and record.levelno >= self._module_log_level[module_name]:
                    return True
        return record.levelno >= self._default_log_level

    @property
    def console_logger(self) -> ConsoleLogger:
        return self._console_logger

    def _set_up(self, echo: bool = False, reinit_db: bool = False) -> None:
        if self._initialized:
            return

        os.environ['TOKENIZERS_PARALLELISM'] = 'false'

        config = Config.get()

        self._initialized = True
        self._media_dir = Config.get().home / 'media'
        self._file_cache_dir = Config.get().home / 'file_cache'
        self._dataset_cache_dir = Config.get().home / 'dataset_cache'
        self._log_dir = Config.get().home / 'logs'
        self._tmp_dir = Config.get().home / 'tmp'

        if not self._media_dir.exists():
            self._media_dir.mkdir()
        if not self._file_cache_dir.exists():
            self._file_cache_dir.mkdir()
        if not self._dataset_cache_dir.exists():
            self._dataset_cache_dir.mkdir()
        if not self._log_dir.exists():
            self._log_dir.mkdir()
        if not self._tmp_dir.exists():
            self._tmp_dir.mkdir()

        self._file_cache_size_g = config.get_float_value('file_cache_size_g')
        if self._file_cache_size_g is None:
            raise excs.Error(
                'pixeltable/file_cache_size_g is missing from configuration\n'
                f'(either add a `file_cache_size_g` entry to the `pixeltable` section of {Config.get().config_file},\n'
                'or set the PIXELTABLE_FILE_CACHE_SIZE_G environment variable)'
            )
        self._pxt_api_key = config.get_string_value('api_key')

        # Disable spurious warnings
        warnings.simplefilter('ignore', category=TqdmWarning)
        if config.get_bool_value('hide_warnings'):
            # Disable more warnings
            warnings.simplefilter('ignore', category=UserWarning)
            warnings.simplefilter('ignore', category=FutureWarning)

        # Set verbose level for user visible console messages
        verbosity = map_level(config.get_int_value('verbosity'))
        stdout_handler = ConsoleOutputHandler(stream=stdout)
        stdout_handler.setLevel(verbosity)
        stdout_handler.addFilter(ConsoleMessageFilter())
        self._logger.addHandler(stdout_handler)
        self._console_logger = ConsoleLogger(self._logger)

        # configure _logger to log to a file
        self._logfilename = datetime.datetime.now().strftime('%Y%m%d_%H%M%S') + '.log'
        fh = logging.FileHandler(self._log_dir / self._logfilename, mode='w')
        fh.setFormatter(logging.Formatter(self._log_fmt_str))
        self._logger.addHandler(fh)

        # configure sqlalchemy logging
        sql_logger = logging.getLogger('sqlalchemy.engine')
        sql_logger.setLevel(logging.INFO)
        sql_logger.addHandler(fh)
        sql_logger.propagate = False

        # configure pyav logging
        av_logfilename = self._logfilename.replace('.log', '_av.log')
        av_fh = logging.FileHandler(self._log_dir / av_logfilename, mode='w')
        av_fh.setFormatter(logging.Formatter(self._log_fmt_str))
        av_logger = logging.getLogger('libav')
        av_logger.addHandler(av_fh)
        av_logger.propagate = False

        # configure web-server logging
        http_logfilename = self._logfilename.replace('.log', '_http.log')
        http_fh = logging.FileHandler(self._log_dir / http_logfilename, mode='w')
        http_fh.setFormatter(logging.Formatter(self._log_fmt_str))
        http_logger = logging.getLogger('pixeltable.http.server')
        http_logger.addHandler(http_fh)
        http_logger.propagate = False

        self.clear_tmp_dir()

        # configure pixeltable database
        self._init_db(config)

        if reinit_db and not self.is_local:
            raise excs.Error(
                'Reinitializing pixeltable database is not supported when running in non-local environment'
            )

        tz_name = config.get_string_value('time_zone')
        if tz_name is not None:
            # Validate tzname
            if not isinstance(tz_name, str):
                self._logger.error('Invalid time zone specified in configuration.')
            else:
                try:
                    _ = ZoneInfo(tz_name)
                except ZoneInfoNotFoundError:
                    self._logger.error(f'Invalid time zone specified in configuration: {tz_name}')

        if reinit_db and self._store_db_exists():
            self._drop_store_db()

        create_db = not self._store_db_exists()

        if create_db:
            self._logger.info(f'creating database at: {self.db_url}')
            self._create_store_db()
        else:
            self._logger.info(f'found database at: {self.db_url}')

        # Create the SQLAlchemy engine. This will also set the default time zone.
        self._create_engine(time_zone_name=tz_name, echo=echo)

        # Create catalog tables and system metadata
        self._init_metadata()

        self.console_logger.info(f'Connected to Pixeltable database at: {self.db_url}')

        # we now have a home directory and db; start other services
        self._set_up_runtime()
        self.log_to_stdout(False)

    def _init_db(self, config: Config) -> None:
        """
        Initialize the pixeltable database along with its associated DBMS.
        """
        db_connect_str = config.get_string_value('DB_CONNECT_STR')
        if db_connect_str is not None:
            try:
                db_url = sql.make_url(db_connect_str)
            except sql.exc.ArgumentError as e:
                error = f'Invalid db connection string {db_connect_str}: {e}'
                self._logger.error(error)
                raise excs.Error(error) from e
            self._db_url = db_url.render_as_string(hide_password=False)
            self._db_name = db_url.database  # use the dbname given in connect string
            dialect = db_url.get_dialect().name
            if dialect == 'cockroachdb':
                self._dbms = CockroachDbms(db_url)
            else:
                raise excs.Error(f'Unsupported DBMS {dialect}')
            # Check if database exists
            if not self._store_db_exists():
                error = f'Database {self._db_name!r} does not exist'
                self._logger.error(error)
                raise excs.Error(error)
            self._logger.info(f'Using database at: {self.db_url}')
        else:
            self._db_name = os.environ.get('PIXELTABLE_DB', 'pixeltable')
            self._pgdata_dir = Path(os.environ.get('PIXELTABLE_PGDATA', str(Config.get().home / 'pgdata')))
            # cleanup_mode=None will leave the postgres process running after Python exits
            # cleanup_mode='stop' will terminate the postgres process when Python exits
            # On Windows, we need cleanup_mode='stop' because child processes are killed automatically when the parent
            # process (such as Terminal or VSCode) exits, potentially leaving it in an unusable state.
            cleanup_mode = 'stop' if platform.system() == 'Windows' else None
            self._db_server = pixeltable_pgserver.get_server(self._pgdata_dir, cleanup_mode=cleanup_mode)
            self._db_url = self._db_server.get_uri(database=self._db_name, driver='psycopg')
            self._dbms = PostgresqlDbms(sql.make_url(self._db_url))
        assert self._dbms is not None
        assert self._db_url is not None
        assert self._db_name is not None

    def _init_metadata(self) -> None:
        """
        Create pixeltable metadata tables and system metadata.
        This is an idempotent operation.
        """
        assert self._sa_engine is not None
        from pixeltable import metadata

        metadata.schema.base_metadata.create_all(self._sa_engine, checkfirst=True)
        metadata.create_system_info(self._sa_engine)

    def _create_engine(self, time_zone_name: Optional[str], echo: bool = False) -> None:
        connect_args = {} if time_zone_name is None else {'options': f'-c timezone={time_zone_name}'}
        self._sa_engine = sql.create_engine(
            self.db_url, echo=echo, isolation_level=self._dbms.transaction_isolation_level, connect_args=connect_args
        )

        self._logger.info(f'Created SQLAlchemy engine at: {self.db_url}')

        with self.engine.begin() as conn:
            tz_name = conn.execute(sql.text('SHOW TIME ZONE')).scalar()
            assert isinstance(tz_name, str)
            self._logger.info(f'Database time zone is now: {tz_name}')
            self._default_time_zone = ZoneInfo(tz_name)

    def _store_db_exists(self) -> bool:
        assert self._db_name is not None
        # don't try to connect to self.db_name, it may not exist
        engine = sql.create_engine(self._dbms.default_system_db_url(), future=True)
        try:
            with engine.begin() as conn:
                stmt = f"SELECT COUNT(*) FROM pg_database WHERE datname = '{self._db_name}'"
                result = conn.scalar(sql.text(stmt))
                assert result <= 1
                return result == 1
        finally:
            engine.dispose()

    def _create_store_db(self) -> None:
        assert self._db_name is not None
        # create the db
        engine = sql.create_engine(self._dbms.default_system_db_url(), future=True, isolation_level='AUTOCOMMIT')
        preparer = engine.dialect.identifier_preparer
        try:
            with engine.begin() as conn:
                stmt = self._dbms.create_db_stmt(preparer.quote(self._db_name))
                conn.execute(sql.text(stmt))
        finally:
            engine.dispose()

        # enable pgvector
        engine = sql.create_engine(self.db_url, future=True, isolation_level='AUTOCOMMIT')
        try:
            with engine.begin() as conn:
                conn.execute(sql.text('CREATE EXTENSION vector'))
        finally:
            engine.dispose()

    def _drop_store_db(self) -> None:
        assert self._db_name is not None
        engine = sql.create_engine(self._dbms.default_system_db_url(), future=True, isolation_level='AUTOCOMMIT')
        preparer = engine.dialect.identifier_preparer
        try:
            with engine.begin() as conn:
                # terminate active connections
                if self._db_server is not None:
                    stmt = f"""
                        SELECT pg_terminate_backend(pg_stat_activity.pid)
                        FROM pg_stat_activity
                        WHERE pg_stat_activity.datname = '{self._db_name}'
                        AND pid <> pg_backend_pid()
                    """
                    conn.execute(sql.text(stmt))
                # drop db
                stmt = self._dbms.drop_db_stmt(preparer.quote(self._db_name))
                conn.execute(sql.text(stmt))
        finally:
            engine.dispose()

    def _upgrade_metadata(self) -> None:
        from pixeltable import metadata

        metadata.upgrade_md(self._sa_engine)

    @property
    def pxt_api_key(self) -> str:
        if self._pxt_api_key is None:
            raise excs.Error(
                'No API key is configured. Set the PIXELTABLE_API_KEY environment variable, or add an entry to '
                'config.toml as described here:\nhttps://pixeltable.github.io/pixeltable/config/'
            )
        return self._pxt_api_key

    def get_client(self, name: str) -> Any:
        """
        Gets the client with the specified name, initializing it if necessary.

        Args:
            - name: The name of the client
        """
        cl = _registered_clients[name]
        if cl.client_obj is not None:
            return cl.client_obj  # Already initialized

        # Construct a client, retrieving each parameter from config.

        init_kwargs: dict[str, str] = {}
        for param in cl.param_names:
            arg = Config.get().get_string_value(param, section=name)
            if arg is not None and len(arg) > 0:
                init_kwargs[param] = arg
            else:
                raise excs.Error(
                    f'`{name}` client not initialized: parameter `{param}` is not configured.\n'
                    f'To fix this, specify the `{name.upper()}_{param.upper()}` environment variable, '
                    f'or put `{param.lower()}` in the `{name.lower()}` section of $PIXELTABLE_HOME/config.toml.'
                )

        cl.client_obj = cl.init_fn(**init_kwargs)
        self._logger.info(f'Initialized `{name}` client.')
        return cl.client_obj

    def _start_web_server(self) -> None:
        """
        The http server root is the file system root.
        eg: /home/media/foo.mp4 is located at http://127.0.0.1:{port}/home/media/foo.mp4
        in windows, the server will translate paths like http://127.0.0.1:{port}/c:/media/foo.mp4
        This arrangement enables serving media hosted within _home,
        as well as external media inserted into pixeltable or produced by pixeltable.
        The port is chosen dynamically to prevent conflicts.
        """
        # Port 0 means OS picks one for us.
        self._httpd = make_server('127.0.0.1', 0)
        port = self._httpd.server_address[1]
        self._http_address = f'http://127.0.0.1:{port}'

        def run_server() -> None:
            logging.log(logging.INFO, f'running web server at {self._http_address}')
            self._httpd.serve_forever()

        # Run the server in a separate thread
        thread = threading.Thread(target=run_server, daemon=True)
        thread.start()

    def _set_up_runtime(self) -> None:
        """Check for and start runtime services"""
        self._start_web_server()
        self.__register_packages()

    def __register_packages(self) -> None:
        """Declare optional packages that are utilized by some parts of the code."""
        self.__register_package('anthropic')
        self.__register_package('boto3')
        self.__register_package('datasets')
        self.__register_package('fiftyone')
        self.__register_package('fireworks', library_name='fireworks-ai')
        self.__register_package('google.generativeai', library_name='google-generativeai')
        self.__register_package('huggingface_hub', library_name='huggingface-hub')
        self.__register_package('label_studio_sdk', library_name='label-studio-sdk')
        self.__register_package('llama_cpp', library_name='llama-cpp-python')
        self.__register_package('mistralai')
        self.__register_package('mistune')
        self.__register_package('ollama')
        self.__register_package('openai')
        self.__register_package('openpyxl')
        self.__register_package('pyarrow')
        self.__register_package('pydantic')
        self.__register_package('replicate')
        self.__register_package('sentencepiece')
        self.__register_package('sentence_transformers', library_name='sentence-transformers')
        self.__register_package('spacy')
        self.__register_package('tiktoken')
        self.__register_package('together')
        self.__register_package('torch')
        self.__register_package('torchaudio')
        self.__register_package('torchvision')
        self.__register_package('transformers')
        self.__register_package('whisper', library_name='openai-whisper')
        self.__register_package('whisperx')
        self.__register_package('yolox', library_name='pixeltable-yolox')

    def __register_package(self, package_name: str, library_name: Optional[str] = None) -> None:
        is_installed: bool
        try:
            is_installed = importlib.util.find_spec(package_name) is not None
        except ModuleNotFoundError:
            # This can happen if the parent of `package_name` is not installed.
            is_installed = False
        self.__optional_packages[package_name] = PackageInfo(
            is_installed=is_installed,
            library_name=library_name or package_name,  # defaults to package_name unless specified otherwise
        )

    def require_package(self, package_name: str, min_version: Optional[list[int]] = None) -> None:
        """
        Checks whether the specified optional package is available. If not, raises an exception
        with an error message informing the user how to install it.
        """
        assert package_name in self.__optional_packages
        package_info = self.__optional_packages[package_name]

        if not package_info.is_installed:
            # Check again whether the package has been installed.
            # We do this so that if a user gets an "optional library not found" error message, they can
            # `pip install` the library and re-run the Pixeltable operation without having to restart
            # their Python session.
            package_info.is_installed = importlib.util.find_spec(package_name) is not None
            if not package_info.is_installed:
                # Still not found.
                raise excs.Error(
                    f'This feature requires the `{package_name}` package. To install it, run: '
                    f'`pip install -U {package_info.library_name}`'
                )

        if min_version is None:
            return

        # check whether we have a version >= the required one
        if package_info.version is None:
            module = importlib.import_module(package_name)
            package_info.version = [int(x) for x in module.__version__.split('.')]

        if min_version > package_info.version:
            raise excs.Error(
                f'The installed version of package `{package_name}` is '
                f'{".".join(str(v) for v in package_info.version)}, '
                f'but version >={".".join(str(v) for v in min_version)} is required. '
                f'To fix this, run: `pip install -U {package_info.library_name}`'
            )

    def clear_tmp_dir(self) -> None:
        for path in glob.glob(f'{self._tmp_dir}/*'):
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

    def num_tmp_files(self) -> int:
        return len(glob.glob(f'{self._tmp_dir}/*'))

    def create_tmp_path(self, extension: str = '') -> Path:
        return self._tmp_dir / f'{uuid.uuid4()}{extension}'

    # def get_resource_pool_info(self, pool_id: str, pool_info_cls: Optional[Type[T]]) -> T:
    def get_resource_pool_info(self, pool_id: str, make_pool_info: Optional[Callable[[], T]] = None) -> T:
        """Returns the info object for the given id, creating it if necessary."""
        info = self._resource_pool_info.get(pool_id)
        if info is None and make_pool_info is not None:
            info = make_pool_info()
            self._resource_pool_info[pool_id] = info
        return info

    @property
    def media_dir(self) -> Path:
        assert self._media_dir is not None
        return self._media_dir

    @property
    def file_cache_dir(self) -> Path:
        assert self._file_cache_dir is not None
        return self._file_cache_dir

    @property
    def dataset_cache_dir(self) -> Path:
        assert self._dataset_cache_dir is not None
        return self._dataset_cache_dir

    @property
    def tmp_dir(self) -> Path:
        assert self._tmp_dir is not None
        return self._tmp_dir

    @property
    def engine(self) -> sql.engine.base.Engine:
        assert self._sa_engine is not None
        return self._sa_engine

    @property
    def spacy_nlp(self) -> spacy.Language:
        Env.get().require_package('spacy')
        if self._spacy_nlp is None:
            self.__init_spacy()
        assert self._spacy_nlp is not None
        return self._spacy_nlp

    def __init_spacy(self) -> None:
        """
        spaCy relies on a pip-installed model to operate. In order to avoid requiring the model as a separate
        dependency, we install it programmatically here. This should cause no problems, since the model packages
        have no sub-dependencies (in fact, this is how spaCy normally manages its model resources).
        """
        import spacy
        from spacy.cli.download import get_model_filename

        spacy_model = 'en_core_web_sm'
        spacy_model_version = '3.7.1'
        filename = get_model_filename(spacy_model, spacy_model_version, sdist=False)
        url = f'{spacy.about.__download_url__}/{filename}'
        # Try to `pip install` the model. We set check=False; if the pip command fails, it's not necessarily
        # a problem, because the model might have been installed on a previous attempt.
        self._logger.info(f'Ensuring spaCy model is installed: {filename}')
        ret = subprocess.run([sys.executable, '-m', 'pip', 'install', '-qU', url], check=False)
        if ret.returncode != 0:
            self._logger.warning(f'pip install failed for spaCy model: {filename}')
        self._logger.info(f'Loading spaCy model: {spacy_model}')
        try:
            self._spacy_nlp = spacy.load(spacy_model)
        except Exception as exc:
            raise excs.Error(f'Failed to load spaCy model: {spacy_model}') from exc


def register_client(name: str) -> Callable:
    """Decorator that registers a third-party API client for use by Pixeltable.

    The decorated function is an initialization wrapper for the client, and can have
    any number of string parameters, with a signature such as:

    ```
    def my_client(api_key: str, url: str) -> my_client_sdk.Client:
        return my_client_sdk.Client(api_key=api_key, url=url)
    ```

    The initialization wrapper will not be called immediately; initialization will
    be deferred until the first time the client is used. At initialization time,
    Pixeltable will attempt to load the client parameters from config. For each
    config parameter:
    - If an environment variable named MY_CLIENT_API_KEY (for example) is set, use it;
    - Otherwise, look for 'api_key' in the 'my_client' section of config.toml.

    If all config parameters are found, Pixeltable calls the initialization function;
    otherwise it throws an exception.

    Args:
        - name (str): The name of the API client (e.g., 'openai' or 'label-studio').
    """

    def decorator(fn: Callable) -> None:
        sig = inspect.signature(fn)
        param_names = list(sig.parameters.keys())
        _registered_clients[name] = ApiClient(init_fn=fn, param_names=param_names)

    return decorator


_registered_clients: dict[str, ApiClient] = {}


@dataclass
class ApiClient:
    init_fn: Callable
    param_names: list[str]
    client_obj: Optional[Any] = None


@dataclass
class PackageInfo:
    is_installed: bool
    library_name: str  # pypi library name (may be different from package name)
    version: Optional[list[int]] = None  # installed version, as a list of components (such as [3,0,2] for "3.0.2")


TIME_FORMAT = '%H:%M.%S %f'


@dataclass
class RateLimitsInfo:
    """
    Abstract base class for resource pools made up of rate limits for different resources.

    Rate limits and currently remaining resources are periodically reported via record().

    Subclasses provide operational customization via:
    - get_retry_delay()
    - get_request_resources(self, ...) -> dict[str, int]
    with parameters that are a subset of those of the udf that creates the subclass's instance
    """

    # get_request_resources:
    # - Returns estimated resources needed for a specific request (ie, a single udf call) as a dict (key: resource name)
    # - parameters are a subset of those of the udf
    # - this is not a class method because the signature depends on the instantiating udf
    get_request_resources: Callable[..., dict[str, int]]

    resource_limits: dict[str, RateLimitInfo] = field(default_factory=dict)

    def is_initialized(self) -> bool:
        return len(self.resource_limits) > 0

    def reset(self) -> None:
        self.resource_limits.clear()

    def record(self, **kwargs: Any) -> None:
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        if len(self.resource_limits) == 0:
            self.resource_limits = {k: RateLimitInfo(k, now, *v) for k, v in kwargs.items() if v is not None}
            # TODO: remove
            for info in self.resource_limits.values():
                _logger.debug(
                    f'Init {info.resource} rate limit: rem={info.remaining} '
                    f'reset={info.reset_at.strftime(TIME_FORMAT)} delta={(info.reset_at - now).total_seconds()}'
                )
        else:
            for k, v in kwargs.items():
                if v is not None:
                    self.resource_limits[k].update(now, *v)

    @abstractmethod
    def get_retry_delay(self, exc: Exception) -> Optional[float]:
        """Returns number of seconds to wait before retry, or None if not retryable"""
        pass


@dataclass
class RateLimitInfo:
    """Container for rate limit-related information for a single resource."""

    resource: str
    recorded_at: datetime.datetime
    limit: int
    remaining: int
    reset_at: datetime.datetime

    def update(self, recorded_at: datetime.datetime, limit: int, remaining: int, reset_at: datetime.datetime) -> None:
        # we always update everything, even though responses may come back out-of-order: we can't use reset_at to
        # determine order, because it doesn't increase monotonically (the reeset duration shortens as output_tokens
        # are freed up - going from max to actual)
        self.recorded_at = recorded_at
        self.limit = limit
        self.remaining = remaining
        reset_delta = reset_at - self.reset_at
        self.reset_at = reset_at
        # TODO: remove
        _logger.debug(
            f'Update {self.resource} rate limit: rem={self.remaining} reset={self.reset_at.strftime(TIME_FORMAT)} '
            f'reset_delta={reset_delta.total_seconds()} recorded_delta={(self.reset_at - recorded_at).total_seconds()}'
        )



import glob
import os
import re
import shutil
import uuid
from collections import defaultdict
from pathlib import Path
from typing import Optional
from uuid import UUID

from pixeltable.env import Env


class MediaStore:
    """
    Utilities to manage media files stored in Env.media_dir

    Media file names are a composite of: table id, column id, version, uuid:
    the table id/column id/version are redundant but useful for identifying all files for a table
    or all files created for a particular version of a table
    """

    pattern = re.compile(r'([0-9a-fA-F]+)_(\d+)_(\d+)_([0-9a-fA-F]+)')  # tbl_id, col_id, version, uuid

    @classmethod
    def prepare_media_path(cls, tbl_id: UUID, col_id: int, version: int, ext: Optional[str] = None) -> Path:
        """
        Construct a new, unique Path name for a persisted media file, and create the parent directory
        for the new Path if it does not already exist. The Path will reside in
        the environment's media_dir.
        """
        id_hex = uuid.uuid4().hex
        parent = Env.get().media_dir / tbl_id.hex / id_hex[:2] / id_hex[:4]
        parent.mkdir(parents=True, exist_ok=True)
        return parent / f'{tbl_id.hex}_{col_id}_{version}_{id_hex}{ext or ""}'

    @classmethod
    def delete(cls, tbl_id: UUID, version: Optional[int] = None) -> None:
        """Delete all files belonging to tbl_id. If version is not None, delete
        only those files belonging to the specified version."""
        assert tbl_id is not None
        if version is None:
            # Remove the entire folder for this table id.
            path = Env.get().media_dir / tbl_id.hex
            if path.exists():
                shutil.rmtree(path)
        else:
            # Remove only the elements for the specified version.
            paths = glob.glob(str(Env.get().media_dir / tbl_id.hex) + f'/**/{tbl_id.hex}_*_{version}_*', recursive=True)
            for p in paths:
                os.remove(p)

    @classmethod
    def count(cls, tbl_id: UUID) -> int:
        """
        Return number of files for given tbl_id.
        """
        paths = glob.glob(str(Env.get().media_dir / tbl_id.hex) + f'/**/{tbl_id.hex}_*', recursive=True)
        return len(paths)

    @classmethod
    def stats(cls) -> list[tuple[UUID, int, int, int]]:
        paths = glob.glob(str(Env.get().media_dir) + '/**', recursive=True)
        # key: (tbl_id, col_id), value: (num_files, size)
        d: dict[tuple[UUID, int], list[int]] = defaultdict(lambda: [0, 0])
        for p in paths:
            if not os.path.isdir(p):
                matched = re.match(cls.pattern, Path(p).name)
                assert matched is not None
                tbl_id, col_id = UUID(hex=matched[1]), int(matched[2])
                file_info = os.stat(p)
                t = d[tbl_id, col_id]
                t[0] += 1
                t[1] += file_info.st_size
        result = [(tbl_id, col_id, num_files, size) for (tbl_id, col_id), (num_files, size) in d.items()]
        result.sort(key=lambda e: e[3], reverse=True)
        return result



# Descriptive notes for each new metadata version. These are stored in a Python dict
# rather than as a comment, so that the existence of a description can be enforced by
# the unit tests when new versions are added.
VERSION_NOTES = {
    31: 'Add table ids to metadata structs',
    30: 'Store default values and constant arguments as literals',
    29: 'Add user and additional_md fields to metadata structs',
    28: 'Enable view creation from DataFrame with select clause',
    27: 'Enable pxt.query parameterization of limit clauses',
    26: 'Rename clip_text and clip_image to clip',
    25: 'Functions with multiple signatures',
    24: 'Added TableMd/IndexMd.indexed_col_tbl_id',
    23: 'DataFrame.from_clause',
    22: 'TableMd/ColumnMd.media_validation',
    21: 'Separate InlineArray and InlineList',
    20: 'Store DB timestamps in UTC',
    19: 'UDF renames; ImageMemberAccess removal',
    18: 'Restructured index metadata',
    17: 'Renamed remotes to external_stores',
    16: 'Query functions; deferred Expr deserialization',
    15: 'Remotes in table metadata',
}



from __future__ import annotations

import dataclasses
import inspect
import logging
import typing
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Optional

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

if TYPE_CHECKING:
    from pixeltable import exprs

_logger = logging.getLogger('pixeltable')


@dataclasses.dataclass
class Parameter:
    name: str
    col_type: Optional[ts.ColumnType]  # None for variable parameters
    kind: inspect._ParameterKind
    # for some reason, this needs to precede is_batched in the dataclass definition,
    # otherwise Python complains that an argument with a default is followed by an argument without a default
    default: Optional['exprs.Literal'] = None  # default value for the parameter
    is_batched: bool = False  # True if the parameter is a batched parameter (eg, Batch[dict])

    def __post_init__(self) -> None:
        from pixeltable import exprs

        if self.default is not None:
            if self.col_type is None:
                raise excs.Error(f'Cannot have a default value for variable parameter {self.name!r}')
            if not isinstance(self.default, exprs.Literal):
                raise excs.Error(f'Default value for parameter {self.name!r} is not a constant')
            if not self.col_type.is_supertype_of(self.default.col_type):
                raise excs.Error(
                    f'Default value for parameter {self.name!r} is not of type {self.col_type!r}: {self.default}'
                )

    def has_default(self) -> bool:
        return self.default is not None

    def as_dict(self) -> dict[str, Any]:
        return {
            'name': self.name,
            'col_type': self.col_type.as_dict() if self.col_type is not None else None,
            'kind': self.kind.name,
            'is_batched': self.is_batched,
            'default': None if self.default is None else self.default.as_dict(),
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Parameter:
        from pixeltable import exprs

        assert d['default'] is None or isinstance(d['default'], dict), d
        default = None if d['default'] is None else exprs.Literal.from_dict(d['default'])
        return cls(
            name=d['name'],
            col_type=ts.ColumnType.from_dict(d['col_type']) if d['col_type'] is not None else None,
            kind=getattr(inspect.Parameter, d['kind']),
            is_batched=d['is_batched'],
            default=default,
        )

    def to_py_param(self) -> inspect.Parameter:
        py_default = self.default.val if self.default is not None else inspect.Parameter.empty
        return inspect.Parameter(self.name, self.kind, default=py_default)

    def __hash__(self) -> int:
        return hash((self.name, self.col_type, self.kind, self.default, self.is_batched))


T = typing.TypeVar('T')
Batch = typing.Annotated[list[T], 'pxt-batch']


class Signature:
    """
    Represents the signature of a Pixeltable function.

    - self.is_batched: return type is a Batch[...] type
    """

    SPECIAL_PARAM_NAMES: ClassVar[list[str]] = ['group_by', 'order_by']

    def __init__(self, return_type: ts.ColumnType, parameters: list[Parameter], is_batched: bool = False):
        assert isinstance(return_type, ts.ColumnType)
        self.return_type = return_type
        self.is_batched = is_batched
        # we rely on the ordering guarantee of dicts in Python >=3.7
        self.parameters = {p.name: p for p in parameters}
        self.parameters_by_pos = parameters.copy()
        self.constant_parameters = [p for p in parameters if not p.is_batched]
        self.batched_parameters = [p for p in parameters if p.is_batched]
        self.required_parameters = [p for p in parameters if not p.has_default()]
        self.py_signature = inspect.Signature([p.to_py_param() for p in self.parameters_by_pos])

    def get_return_type(self) -> ts.ColumnType:
        assert isinstance(self.return_type, ts.ColumnType)
        return self.return_type

    def as_dict(self) -> dict[str, Any]:
        result = {
            'return_type': self.get_return_type().as_dict(),
            'parameters': [p.as_dict() for p in self.parameters.values()],
            'is_batched': self.is_batched,
        }
        return result

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Signature:
        parameters = [Parameter.from_dict(param_dict) for param_dict in d['parameters']]
        return cls(ts.ColumnType.from_dict(d['return_type']), parameters, d['is_batched'])

    def is_consistent_with(self, other: Signature) -> bool:
        """
        Returns True if this signature is consistent with the other signature.
        S is consistent with T if we could safely replace S by T in any call where S is used. Specifically:
        (i) S.return_type is a supertype of T.return_type
        (ii) For each parameter p in S, there is a parameter q in T such that:
            - p and q have the same name and kind
            - q.col_type is a supertype of p.col_type
        (iii) For each *required* parameter q in T, there is a parameter p in S with the same name (in which
            case the kinds and types must also match, by condition (ii)).
        """
        # Check (i)
        if not self.get_return_type().is_supertype_of(other.get_return_type(), ignore_nullable=True):
            return False

        # Check (ii)
        for param_name, param in self.parameters.items():
            if param_name not in other.parameters:
                return False
            other_param = other.parameters[param_name]
            if (
                param.kind != other_param.kind
                or (param.col_type is None) != (other_param.col_type is None)  # this can happen if they are varargs
                or (
                    param.col_type is not None
                    and not other_param.col_type.is_supertype_of(param.col_type, ignore_nullable=True)
                )
            ):
                return False

        # Check (iii)
        for other_param in other.required_parameters:  # noqa: SIM110
            if other_param.name not in self.parameters:
                return False

        return True

    def validate_args(self, bound_args: dict[str, Optional['exprs.Expr']], context: str = '') -> None:
        if context:
            context = f' ({context})'

        for param_name, arg in bound_args.items():
            assert param_name in self.parameters
            param = self.parameters[param_name]
            is_var_param = param.kind in {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}
            if is_var_param:
                continue
            assert param.col_type is not None

            if arg is None:
                raise excs.Error(f'Parameter {param_name!r}{context}: invalid argument')

            # Check that the argument is consistent with the expected parameter type, with the allowance that
            # non-nullable parameters can still accept nullable arguments (since in that event, FunctionCall.eval()
            # detects the Nones and skips evaluation).
            if not (
                param.col_type.is_supertype_of(arg.col_type, ignore_nullable=True)
                # TODO: this is a hack to allow JSON columns to be passed to functions that accept scalar
                # types. It's necessary to avoid littering notebooks with `apply(str)` calls or equivalent.
                # (Previously, this wasn't necessary because `is_supertype_of()` was improperly implemented.)
                # We need to think through the right way to handle this scenario.
                or (arg.col_type.is_json_type() and param.col_type.is_scalar_type())
            ):
                raise excs.Error(
                    f'Parameter {param_name!r}{context}: argument type {arg.col_type} does not'
                    f' match parameter type {param.col_type}'
                )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Signature):
            return False
        if self.get_return_type() != other.get_return_type():
            return False
        if len(self.parameters) != len(other.parameters):
            return False
        # ignore the parameter name
        for param, other_param in zip(self.parameters.values(), other.parameters.values()):
            if param.col_type != other_param.col_type or param.kind != other_param.kind:
                return False
        return True

    def __hash__(self) -> int:
        return hash((self.return_type, self.parameters))

    def __str__(self) -> str:
        param_strs: list[str] = []
        for p in self.parameters.values():
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                param_strs.append(f'*{p.name}')
            elif p.kind == inspect.Parameter.VAR_KEYWORD:
                param_strs.append(f'**{p.name}')
            else:
                param_strs.append(f'{p.name}: {p.col_type}')
        return f'({", ".join(param_strs)}) -> {self.get_return_type()}'

    @classmethod
    def _infer_type(cls, annotation: Optional[type]) -> tuple[Optional[ts.ColumnType], Optional[bool]]:
        """Returns: (column type, is_batched) or (None, ...) if the type cannot be inferred"""
        if annotation is None:
            return (None, None)
        py_type: Optional[type] = None
        is_batched = False
        if typing.get_origin(annotation) == typing.Annotated:
            type_args = typing.get_args(annotation)
            if len(type_args) == 2 and type_args[1] == 'pxt-batch':
                # this is our Batch
                assert typing.get_origin(type_args[0]) is list
                is_batched = True
                py_type = typing.get_args(type_args[0])[0]
        if py_type is None:
            py_type = annotation
        col_type = ts.ColumnType.from_python_type(py_type)
        return (col_type, is_batched)

    @classmethod
    def create_parameters(
        cls,
        py_fn: Optional[Callable] = None,
        py_params: Optional[list[inspect.Parameter]] = None,
        param_types: Optional[list[ts.ColumnType]] = None,
        type_substitutions: Optional[dict] = None,
        is_cls_method: bool = False,
    ) -> list[Parameter]:
        from pixeltable import exprs

        assert (py_fn is None) != (py_params is None)
        if py_fn is not None:
            sig = inspect.signature(py_fn)
            py_params = list(sig.parameters.values())
        parameters: list[Parameter] = []

        if type_substitutions is None:
            type_substitutions = {}

        for idx, param in enumerate(py_params):
            if is_cls_method and idx == 0:
                continue  # skip 'self' or 'cls' parameter
            if param.name in cls.SPECIAL_PARAM_NAMES:
                raise excs.Error(f'{param.name!r} is a reserved parameter name')
            if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
                parameters.append(Parameter(param.name, col_type=None, kind=param.kind))
                continue

            # check non-var parameters for name collisions and default value compatibility
            if param_types is not None:
                if idx >= len(param_types):
                    raise excs.Error(f'Missing type for parameter {param.name!r}')
                param_type = param_types[idx]
                is_batched = False
            else:
                # Look up the substitution for param.annotation, defaulting to param.annotation if there is none
                py_type = type_substitutions.get(param.annotation, param.annotation)
                param_type, is_batched = cls._infer_type(py_type)
                if param_type is None:
                    raise excs.Error(f'Cannot infer pixeltable type for parameter {param.name!r}')

            default = None if param.default is inspect.Parameter.empty else exprs.Expr.from_object(param.default)
            if not (default is None or isinstance(default, exprs.Literal)):
                raise excs.Error(f'Default value for parameter {param.name!r} must be a constant')

            parameters.append(
                Parameter(param.name, col_type=param_type, kind=param.kind, is_batched=is_batched, default=default)
            )

        return parameters

    @classmethod
    def create(
        cls,
        py_fn: Callable,
        param_types: Optional[list[ts.ColumnType]] = None,
        return_type: Optional[ts.ColumnType] = None,
        type_substitutions: Optional[dict] = None,
        is_cls_method: bool = False,
    ) -> Signature:
        """Create a signature for the given Callable.
        Infer the parameter and return types, if none are specified.
        Raises an exception if the types cannot be inferred.
        """
        if type_substitutions is None:
            type_substitutions = {}

        parameters = cls.create_parameters(
            py_fn=py_fn, param_types=param_types, is_cls_method=is_cls_method, type_substitutions=type_substitutions
        )
        sig = inspect.signature(py_fn)
        if return_type is None:
            # Look up the substitution for sig.return_annotation, defaulting to return_annotation if there is none
            py_type = type_substitutions.get(sig.return_annotation, sig.return_annotation)
            return_type, return_is_batched = cls._infer_type(py_type)
            if return_type is None:
                raise excs.Error('Cannot infer pixeltable return type')
        else:
            _, return_is_batched = cls._infer_type(sig.return_annotation)

        return Signature(return_type, parameters, return_is_batched)



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Replicate API. In order to use them, you must
first `pip install replicate` and configure your Replicate credentials, as described in
the [Working with Replicate](https://pixeltable.readme.io/docs/working-with-replicate) tutorial.
"""

from typing import TYPE_CHECKING, Any

import pixeltable as pxt
from pixeltable.env import Env, register_client
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import replicate  # type: ignore[import-untyped]


@register_client('replicate')
def _(api_token: str) -> 'replicate.Client':
    import replicate

    return replicate.Client(api_token=api_token)


def _replicate_client() -> 'replicate.Client':
    return Env.get().get_client('replicate')


@pxt.udf(resource_pool='request-rate:replicate')
async def run(input: dict[str, Any], *, ref: str) -> dict[str, Any]:
    """
    Run a model on Replicate.

    For additional details, see: <https://replicate.com/docs/topics/models/run-a-model>

    Request throttling:
    Applies the rate limit set in the config (section `replicate`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install replicate`

    Args:
        input: The input parameters for the model.
        ref: The name of the model to run.

    Returns:
        The output of the model.

    Examples:
        Add a computed column that applies the model `meta/meta-llama-3-8b-instruct`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> input = {'system_prompt': 'You are a helpful assistant.', 'prompt': tbl.prompt}
        ... tbl.add_computed_column(response=run(input, ref='meta/meta-llama-3-8b-instruct'))

        Add a computed column that uses the model `black-forest-labs/flux-schnell`
        to generate images from an existing Pixeltable column `tbl.prompt`:

        >>> input = {'prompt': tbl.prompt, 'go_fast': True, 'megapixels': '1'}
        ... tbl.add_computed_column(response=run(input, ref='black-forest-labs/flux-schnell'))
        ... tbl.add_computed_column(image=tbl.response.output[0].astype(pxt.Image))
    """
    Env.get().require_package('replicate')
    return await _replicate_client().async_run(ref, input, use_file_output=False)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

from typing import Any, Optional

import sqlalchemy as sql

from pixeltable import exceptions as excs, type_system as ts

from .data_row import DataRow
from .expr import Expr
from .globals import ArithmeticOperator
from .literal import Literal
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class ArithmeticExpr(Expr):
    """
    Allows arithmetic exprs on json paths
    """

    operator: ArithmeticOperator

    def __init__(self, operator: ArithmeticOperator, op1: Expr, op2: Expr):
        if op1.col_type.is_json_type() or op2.col_type.is_json_type() or operator == ArithmeticOperator.DIV:
            # we assume it's a float
            super().__init__(ts.FloatType(nullable=(op1.col_type.nullable or op2.col_type.nullable)))
        else:
            super().__init__(op1.col_type.supertype(op2.col_type))
        self.operator = operator
        self.components = [op1, op2]

        # do typechecking after initialization in order for __str__() to work
        if not op1.col_type.is_numeric_type() and not op1.col_type.is_json_type():
            raise excs.Error(f'{self}: {operator} requires numeric types, but {op1} has type {op1.col_type}')
        if not op2.col_type.is_numeric_type() and not op2.col_type.is_json_type():
            raise excs.Error(f'{self}: {operator} requires numeric types, but {op2} has type {op2.col_type}')

        self.id = self._create_id()

    @property
    def _op1(self) -> Expr:
        return self.components[0]

    @property
    def _op2(self) -> Expr:
        return self.components[1]

    def __repr__(self) -> str:
        # add parentheses around operands that are ArithmeticExprs to express precedence
        op1_str = f'({self._op1})' if isinstance(self._op1, ArithmeticExpr) else str(self._op1)
        op2_str = f'({self._op2})' if isinstance(self._op2, ArithmeticExpr) else str(self._op2)
        return f'{op1_str} {self.operator} {op2_str}'

    def _equals(self, other: ArithmeticExpr) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('operator', self.operator.value)]

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        assert self.col_type.is_int_type() or self.col_type.is_float_type() or self.col_type.is_json_type()
        left = sql_elements.get(self._op1)
        right = sql_elements.get(self._op2)
        if left is None or right is None:
            return None
        if self.operator == ArithmeticOperator.ADD:
            return left + right
        if self.operator == ArithmeticOperator.SUB:
            return left - right
        if self.operator == ArithmeticOperator.MUL:
            return left * right
        if self.operator == ArithmeticOperator.DIV:
            assert self.col_type.is_float_type()
            # Avoid DivisionByZero: if right is 0, make this a NULL
            # TODO: Should we cast the NULLs to NaNs when they are retrieved back into Python?
            nullif = sql.sql.func.nullif(right, 0)
            # We have to cast to a `float`, or else we'll get a `Decimal`
            return sql.sql.expression.cast(left / nullif, self.col_type.to_sa_type())
        if self.operator == ArithmeticOperator.MOD:
            if self.col_type.is_int_type():
                nullif = sql.sql.func.nullif(right, 0)
                return left % nullif
            if self.col_type.is_float_type():
                # Postgres does not support modulus for floats
                return None
        if self.operator == ArithmeticOperator.FLOORDIV:
            # Postgres has a DIV operator, but it behaves differently from Python's // operator
            # (Postgres rounds toward 0, Python rounds toward negative infinity)
            # We need the behavior to be consistent, so that expressions will evaluate the same way
            # whether or not their operands can be translated to SQL. These SQL clauses should
            # mimic the behavior of Python's // operator.
            nullif = sql.sql.func.nullif(right, 0)
            if self.col_type.is_int_type():
                return sql.sql.expression.cast(sql.func.floor(left / nullif), self.col_type.to_sa_type())
            if self.col_type.is_float_type():
                return sql.sql.expression.cast(sql.func.floor(left / nullif), self.col_type.to_sa_type())
        raise AssertionError()

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        op1_val = data_row[self._op1.slot_idx]
        op2_val = data_row[self._op2.slot_idx]

        # if one or both columns is JsonTyped, we need a dynamic check that they are numeric
        if self._op1.col_type.is_json_type() and op1_val is not None and not isinstance(op1_val, (int, float)):
            raise excs.Error(
                f'{self.operator} requires numeric types, but {self._op1} has type {type(op1_val).__name__}'
            )
        if self._op2.col_type.is_json_type() and op2_val is not None and not isinstance(op2_val, (int, float)):
            raise excs.Error(
                f'{self.operator} requires numeric types, but {self._op2} has type {type(op2_val).__name__}'
            )

        data_row[self.slot_idx] = self.eval_nullable(op1_val, op2_val)

    def eval_nullable(self, op1_val: Optional[float], op2_val: Optional[float]) -> Optional[float]:
        """
        Return the result of evaluating the expression on two nullable int/float operands,
        None is interpreted as SQL NULL
        """
        if op1_val is None or op2_val is None:
            return None
        return self.eval_non_null(op1_val, op2_val)

    def eval_non_null(self, op1_val: float, op2_val: float) -> float:
        """
        Return the result of evaluating the expression on two int/float operands
        """
        if self.operator == ArithmeticOperator.ADD:
            return op1_val + op2_val
        elif self.operator == ArithmeticOperator.SUB:
            return op1_val - op2_val
        elif self.operator == ArithmeticOperator.MUL:
            return op1_val * op2_val
        elif self.operator == ArithmeticOperator.DIV:
            return op1_val / op2_val
        elif self.operator == ArithmeticOperator.MOD:
            return op1_val % op2_val
        elif self.operator == ArithmeticOperator.FLOORDIV:
            return op1_val // op2_val

    def as_literal(self) -> Optional[Literal]:
        op1_lit = self._op1.as_literal()
        if op1_lit is None:
            return None
        op2_lit = self._op2.as_literal()
        if op2_lit is None:
            return None
        op1_val = op1_lit.val
        assert op1_lit.col_type.is_numeric_type() or op1_val is None
        op2_val = op2_lit.val
        assert op2_lit.col_type.is_numeric_type() or op2_val is None

        return Literal(self.eval_nullable(op1_val, op2_val), self.col_type)  # type: ignore[arg-type]

    def _as_dict(self) -> dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> ArithmeticExpr:
        assert 'operator' in d
        assert len(components) == 2
        return cls(ArithmeticOperator(d['operator']), components[0], components[1])



from typing import TYPE_CHECKING, Optional

import sqlalchemy as sql

# TODO: why does this import result in a circular import, but the one im embedding_index.py doesn't?
# import pixeltable.catalog as catalog
import pixeltable.exceptions as excs
from pixeltable import catalog, exprs
from pixeltable.env import Env
from pixeltable.func.udf import udf

from .base import IndexBase

if TYPE_CHECKING:
    import pixeltable.exprs


class BtreeIndex(IndexBase):
    """
    Interface to B-tree indices in Postgres.
    """

    MAX_STRING_LEN = 256

    value_expr: 'pixeltable.exprs.Expr'

    @staticmethod
    @udf
    def str_filter(s: Optional[str]) -> Optional[str]:
        if s is None:
            return None
        return s[: BtreeIndex.MAX_STRING_LEN]

    def __init__(self, c: 'catalog.Column'):
        if not c.col_type.is_scalar_type() and not c.col_type.is_media_type():
            raise excs.Error(f'Index on column {c.name}: B-tree index requires scalar or media type, got {c.col_type}')
        if c.col_type.is_media_type():
            # an index on a media column is an index on the file url
            # no validation for media columns: we're only interested in the string value
            self.value_expr = exprs.ColumnRef(c, perform_validation=False)
        else:
            self.value_expr = (
                BtreeIndex.str_filter(exprs.ColumnRef(c)) if c.col_type.is_string_type() else exprs.ColumnRef(c)
            )

    def index_value_expr(self) -> 'exprs.Expr':
        return self.value_expr

    def records_value_errors(self) -> bool:
        return False

    def index_sa_type(self) -> sql.types.TypeEngine:
        """Return the sqlalchemy type of the index value column"""
        return self.value_expr.col_type.to_sa_type()

    def create_index(self, index_name: str, index_value_col: 'catalog.Column') -> None:
        """Create the index on the index value column"""
        idx = sql.Index(index_name, index_value_col.sa_col, postgresql_using='btree')
        conn = Env.get().conn
        idx.create(bind=conn)

    @classmethod
    def display_name(cls) -> str:
        return 'btree'

    def as_dict(self) -> dict:
        return {}

    @classmethod
    def from_dict(cls, c: 'catalog.Column', d: dict) -> 'BtreeIndex':
        return cls(c)



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=18)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    # Migrate a few changed function names
    if k == 'path' and v == 'pixeltable.functions.string.str_format':
        return 'path', 'pixeltable.functions.string.format'
    if k == 'path' and v.startswith('pixeltable.functions.pil.image'):
        return 'path', v.replace('pixeltable.functions.pil.image', 'pixeltable.functions.image')
    # Migrate deprecated `ImageMemberAccess` expressions to `FunctionCall`s
    if isinstance(v, dict) and '_classname' in v and v['_classname'] == 'ImageMemberAccess':
        member_name = v['member_name']
        new_v = {
            'fn': {
                'path': f'pixeltable.functions.image.{member_name}',
                '_classpath': 'pixeltable.func.callable_function.CallableFunction',
            },
            'args': [[0, None]],
            'kwargs': {},
            '_classname': 'FunctionCall',
            'components': v['components'],
            'group_by_stop_idx': 0,
            'group_by_start_idx': 0,
            'order_by_start_idx': 1,
        }
        return k, new_v
    return None



from __future__ import annotations

from typing import Any, Optional, cast
from uuid import UUID

import sqlalchemy as sql

from pixeltable import catalog, type_system as ts

from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class RowidRef(Expr):
    """A reference to a part of a table rowid

    This is used internally to support grouping by a base table and for references to the 'pos' column.
    When a RowidRef is part of a computed column in a view, the view's TableVersion isn't available when
    _from_dict()/init() is called, which is why this class effectively has two separate paths for construction
    (with and without a TableVersion).
    """

    tbl: Optional[catalog.TableVersionHandle]
    normalized_base: Optional[catalog.TableVersionHandle]
    tbl_id: UUID
    normalized_base_id: UUID
    rowid_component_idx: int

    def __init__(
        self,
        tbl: catalog.TableVersionHandle,
        idx: int,
        tbl_id: Optional[UUID] = None,
        normalized_base_id: Optional[UUID] = None,
    ):
        super().__init__(ts.IntType(nullable=False))
        self.tbl = tbl
        if tbl is not None:
            # normalize to simplify comparisons: we refer to the lowest base table that has the requested rowid idx
            # (which has the same values as all its descendent views)
            normalized_base = tbl
            # don't try to reference tbl.store_tbl here
            while normalized_base.get().base is not None and normalized_base.get().base.get().num_rowid_columns() > idx:
                normalized_base = normalized_base.get().base
            self.normalized_base = normalized_base
        else:
            self.normalized_base = None

        # if we're initialized by _from_dict(), we only have the ids, not the TableVersion itself
        self.tbl_id = tbl.id if tbl is not None else tbl_id
        self.normalized_base_id = self.normalized_base.id if self.normalized_base is not None else normalized_base_id
        self.rowid_component_idx = idx
        self.id = self._create_id()

    def default_column_name(self) -> Optional[str]:
        return str(self)

    def _equals(self, other: RowidRef) -> bool:
        return (
            self.normalized_base_id == other.normalized_base_id
            and self.rowid_component_idx == other.rowid_component_idx
        )

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [
            *super()._id_attrs(),
            ('normalized_base_id', self.normalized_base_id),
            ('idx', self.rowid_component_idx),
        ]

    def __repr__(self) -> str:
        # check if this is the pos column of a component view
        from pixeltable import store

        tbl = self.tbl.get() if self.tbl is not None else catalog.Catalog.get().get_tbl_version(self.tbl_id, None)
        if (
            tbl.is_component_view
            and self.rowid_component_idx == cast(store.StoreComponentView, tbl.store_tbl).pos_col_idx
        ):
            return catalog.globals._POS_COLUMN_NAME
        return ''

    def set_tbl(self, tbl: catalog.TableVersionPath) -> None:
        """Change the table that is being referenced.
        This can be necessary during query planning, because at that stage we try to minimize the total number of
        tables that are referenced/need to be joined.
        We can only change to a view of the original table (which shares the base's rowid columns).
        """
        if self.tbl_id == tbl.tbl_version.id:
            return
        base_ids = [tbl_version.id for tbl_version in tbl.get_tbl_versions()]
        assert self.tbl_id in base_ids  # our current TableVersion is a base of the new TableVersion
        self.tbl = tbl.tbl_version
        self.tbl_id = self.tbl.id

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        tbl = self.tbl.get() if self.tbl is not None else catalog.Catalog.get().get_tbl_version(self.tbl_id, None)
        rowid_cols = tbl.store_tbl.rowid_columns()
        return rowid_cols[self.rowid_component_idx]

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        data_row[self.slot_idx] = data_row.pk[self.rowid_component_idx]

    def _as_dict(self) -> dict:
        return {
            'tbl_id': str(self.tbl_id),
            'normalized_base_id': str(self.normalized_base_id),
            'idx': self.rowid_component_idx,
        }

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> RowidRef:
        tbl_id, normalized_base_id, idx = UUID(d['tbl_id']), UUID(d['normalized_base_id']), d['idx']
        return cls(tbl=None, idx=idx, tbl_id=tbl_id, normalized_base_id=normalized_base_id)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional
from uuid import UUID

from .table_version import TableVersion

if TYPE_CHECKING:
    pass

_logger = logging.getLogger('pixeltable')


class TableVersionHandle:
    """
    Indirection mechanism for TableVersion instances, which get resolved against the catalog at runtime.
    """

    id: UUID
    effective_version: Optional[int]
    _tbl_version: Optional[TableVersion]

    def __init__(self, tbl_id: UUID, effective_version: Optional[int], tbl_version: Optional[TableVersion] = None):
        self.id = tbl_id
        self.effective_version = effective_version
        self._tbl_version = tbl_version

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TableVersionHandle):
            return False
        return self.id == other.id and self.effective_version == other.effective_version

    def __hash__(self) -> int:
        return hash((self.id, self.effective_version))

    @classmethod
    def create(cls, tbl_version: TableVersion) -> TableVersionHandle:
        return cls(tbl_version.id, tbl_version.effective_version, tbl_version)

    def get(self) -> TableVersion:
        from .catalog import Catalog

        if self._tbl_version is None:
            self._tbl_version = Catalog.get().get_tbl_version(self.id, self.effective_version)
        return self._tbl_version

    def as_dict(self) -> dict:
        return {'id': str(self.id), 'effective_version': self.effective_version}

    @classmethod
    def from_dict(cls, d: dict) -> TableVersionHandle:
        return cls(UUID(d['id']), d['effective_version'])



from __future__ import annotations

import typing
from typing import Any, Optional, Union

import pixeltable as pxt
import pixeltable.type_system as ts

if typing.TYPE_CHECKING:
    import datasets  # type: ignore[import-untyped]


# note, there are many more types. we allow overrides in the schema_overrides parameter
# to handle cases where the appropriate type is not yet mapped, or to override this mapping.
# https://huggingface.co/docs/datasets/v2.17.0/en/package_reference/main_classes#datasets.Value
_hf_to_pxt: dict[str, ts.ColumnType] = {
    'bool': ts.BoolType(nullable=True),
    'int8': ts.IntType(nullable=True),
    'int16': ts.IntType(nullable=True),
    'int32': ts.IntType(nullable=True),
    'int64': ts.IntType(nullable=True),
    'uint8': ts.IntType(nullable=True),
    'uint16': ts.IntType(nullable=True),
    'uint32': ts.IntType(nullable=True),
    'uint64': ts.IntType(nullable=True),
    'float16': ts.FloatType(nullable=True),
    'float32': ts.FloatType(nullable=True),
    'float64': ts.FloatType(nullable=True),
    'string': ts.StringType(nullable=True),
    'large_string': ts.StringType(nullable=True),
    'timestamp[s]': ts.TimestampType(nullable=True),
    'timestamp[ms]': ts.TimestampType(nullable=True),  # HF dataset iterator converts timestamps to datetime.datetime
    'timestamp[us]': ts.TimestampType(nullable=True),
    'date32': ts.StringType(nullable=True),  # date32 is not supported in pixeltable, use string
    'date64': ts.StringType(nullable=True),  # date64 is not supported in pixeltable, use string
}


def _to_pixeltable_type(feature_type: Any, nullable: bool) -> Optional[ts.ColumnType]:
    """Convert a huggingface feature type to a pixeltable ColumnType if one is defined."""
    import datasets

    if isinstance(feature_type, datasets.ClassLabel):
        # enum, example: ClassLabel(names=['neg', 'pos'], id=None)
        return ts.StringType(nullable=nullable)
    elif isinstance(feature_type, datasets.Value):
        # example: Value(dtype='int64', id=None)
        pt = _hf_to_pxt.get(feature_type.dtype, None)
        return pt.copy(nullable=nullable) if pt is not None else None
    elif isinstance(feature_type, datasets.Sequence):
        # example: cohere wiki. Sequence(feature=Value(dtype='float32', id=None), length=-1, id=None)
        dtype = _to_pixeltable_type(feature_type.feature, nullable)
        length = feature_type.length if feature_type.length != -1 else None
        return ts.ArrayType(shape=(length,), dtype=dtype)
    elif isinstance(feature_type, datasets.Image):
        return ts.ImageType(nullable=nullable)
    else:
        return None


def _get_hf_schema(dataset: Union[datasets.Dataset, datasets.DatasetDict]) -> datasets.Features:
    """Get the schema of a huggingface dataset as a dictionary."""
    import datasets

    first_dataset = dataset if isinstance(dataset, datasets.Dataset) else next(iter(dataset.values()))
    return first_dataset.features


def huggingface_schema_to_pxt_schema(
    hf_schema: datasets.Features, schema_overrides: dict[str, Any], primary_key: list[str]
) -> dict[str, Optional[ts.ColumnType]]:
    """Generate a pixeltable schema from a huggingface dataset schema.
    Columns without a known mapping are mapped to None
    """
    pixeltable_schema = {
        column_name: _to_pixeltable_type(feature_type, column_name not in primary_key)
        if column_name not in schema_overrides
        else schema_overrides[column_name]
        for column_name, feature_type in hf_schema.items()
    }
    return pixeltable_schema


def import_huggingface_dataset(
    table_path: str,
    dataset: Union[datasets.Dataset, datasets.DatasetDict],
    *,
    schema_overrides: Optional[dict[str, Any]] = None,
    primary_key: Optional[Union[str, list[str]]] = None,
    **kwargs: Any,
) -> pxt.Table:
    """Create a new base table from a Huggingface dataset, or dataset dict with multiple splits.
        Requires `datasets` library to be installed.

    Args:
        table_path: Path to the table.
        dataset: Huggingface [`datasets.Dataset`](https://huggingface.co/docs/datasets/en/package_reference/main_classes#datasets.Dataset)
            or [`datasets.DatasetDict`](https://huggingface.co/docs/datasets/en/package_reference/main_classes#datasets.DatasetDict)
            to insert into the table.
        schema_overrides: If specified, then for each (name, type) pair in `schema_overrides`, the column with
            name `name` will be given type `type`, instead of being inferred from the `Dataset` or `DatasetDict`.
            The keys in `schema_overrides` should be the column names of the `Dataset` or `DatasetDict` (whether or not
            they are valid Pixeltable identifiers).
        primary_key: The primary key of the table (see [`create_table()`][pixeltable.create_table]).
        kwargs: Additional arguments to pass to `create_table`.
            An argument of `column_name_for_split` must be provided if the source is a DatasetDict.
            This column name will contain the split information. If None, no split information will be stored.

    Returns:
        A handle to the newly created [`Table`][pixeltable.Table].
    """
    return pxt.create_table(
        table_path, source=dataset, schema_overrides=schema_overrides, primary_key=primary_key, extra_args=kwargs
    )



from __future__ import annotations

import asyncio
import datetime
import itertools
import logging
import sys
from typing import Any, Callable, Iterator, Optional, cast

from pixeltable import exprs, func

from .globals import Dispatcher, Evaluator, ExecCtx, FnCallArgs

_logger = logging.getLogger('pixeltable')


class DefaultExprEvaluator(Evaluator):
    """
    Standard expression evaluation using Expr.eval().

    Creates one task per set of rows handed to schedule().

    TODO:
    - parallelize via Ray
    """

    e: exprs.Expr

    def __init__(self, e: exprs.Expr, dispatcher: Dispatcher, exec_ctx: ExecCtx):
        super().__init__(dispatcher, exec_ctx)
        self.e = e

    def schedule(self, rows: list[exprs.DataRow], slot_idx: int) -> None:
        assert self.e.slot_idx >= 0
        task = asyncio.create_task(self.eval(rows))
        self.dispatcher.register_task(task)

    async def eval(self, rows: list[exprs.DataRow]) -> None:
        rows_with_excs: set[int] = set()  # records idxs into rows
        for idx, row in enumerate(rows):
            assert not row.has_val[self.e.slot_idx] and not row.has_exc(self.e.slot_idx)
            if asyncio.current_task().cancelled() or self.dispatcher.exc_event.is_set():
                return
            try:
                self.e.eval(row, self.dispatcher.row_builder)
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                row.set_exc(self.e.slot_idx, exc)
                rows_with_excs.add(idx)
                self.dispatcher.dispatch_exc([row], self.e.slot_idx, exc_tb, self.exec_ctx)
        self.dispatcher.dispatch([rows[i] for i in range(len(rows)) if i not in rows_with_excs], self.exec_ctx)


class FnCallEvaluator(Evaluator):
    """
    Evaluates function calls:
    - batched functions (sync and async): one task per batch
    - async functions: one task per row
    - the rest: one task per set of rows handed to schedule()

    TODO:
    - adaptive batching: finding the optimal batch size based on observed execution times
    """

    fn_call: exprs.FunctionCall
    fn: func.CallableFunction
    scalar_py_fn: Optional[Callable]  # only set for non-batching CallableFunctions

    # only set if fn.is_batched
    call_args_queue: Optional[asyncio.Queue[FnCallArgs]]  # FnCallArgs waiting for execution
    batch_size: Optional[int]

    def __init__(self, fn_call: exprs.FunctionCall, dispatcher: Dispatcher, exec_ctx: ExecCtx):
        super().__init__(dispatcher, exec_ctx)
        self.fn_call = fn_call
        self.fn = cast(func.CallableFunction, fn_call.fn)
        if isinstance(self.fn, func.CallableFunction) and self.fn.is_batched:
            self.call_args_queue = asyncio.Queue[FnCallArgs]()
            # we're not supplying sample arguments there, they're ignored anyway
            self.batch_size = self.fn.get_batch_size()
            self.scalar_py_fn = None
        else:
            self.call_args_queue = None
            self.batch_size = None
            if isinstance(self.fn, func.CallableFunction):
                self.scalar_py_fn = self.fn.py_fn
            else:
                self.scalar_py_fn = None

    def schedule(self, rows: list[exprs.DataRow], slot_idx: int) -> None:
        assert self.fn_call.slot_idx >= 0

        # create FnCallArgs for incoming rows
        skip_rows: list[exprs.DataRow] = []  # skip rows with Nones in non-nullable parameters
        rows_call_args: list[FnCallArgs] = []
        for row in rows:
            args_kwargs = self.fn_call.make_args(row)
            if args_kwargs is None:
                # nothing to do here
                row[self.fn_call.slot_idx] = None
                skip_rows.append(row)
            else:
                args, kwargs = args_kwargs
                rows_call_args.append(FnCallArgs(self.fn_call, [row], args=args, kwargs=kwargs))

        if len(skip_rows) > 0:
            self.dispatcher.dispatch(skip_rows, self.exec_ctx)

        if self.batch_size is not None:
            if not self.is_closed and (len(rows_call_args) + self.call_args_queue.qsize() < self.batch_size):
                # we don't have enough FnCallArgs for a batch, so add them to the queue
                for item in rows_call_args:
                    self.call_args_queue.put_nowait(item)
                return

            # create one task per batch
            combined_call_args = itertools.chain(self._queued_call_args_iter(), rows_call_args)
            while True:
                call_args_batch = list(itertools.islice(combined_call_args, self.batch_size))
                if len(call_args_batch) == 0:
                    break
                if len(call_args_batch) < self.batch_size and not self.is_closed:
                    # we don't have a full batch left: return the rest to the queue
                    assert self.call_args_queue.empty()  # we saw all queued items
                    for item in call_args_batch:
                        self.call_args_queue.put_nowait(item)
                    return

                # turn call_args_batch into a single batched FnCallArgs
                _logger.debug(f'Creating batch of size {len(call_args_batch)} for slot {slot_idx}')
                batched_call_args = self._create_batch_call_args(call_args_batch)
                if self.fn_call.resource_pool is not None:
                    # hand the call off to the resource pool's scheduler
                    scheduler = self.dispatcher.schedulers[self.fn_call.resource_pool]
                    scheduler.submit(batched_call_args, self.exec_ctx)
                else:
                    task = asyncio.create_task(self.eval_batch(batched_call_args))
                    self.dispatcher.register_task(task)

        elif self.fn.is_async:
            if self.fn_call.resource_pool is not None:
                # hand the call off to the resource pool's scheduler
                scheduler = self.dispatcher.schedulers[self.fn_call.resource_pool]
                for item in rows_call_args:
                    scheduler.submit(item, self.exec_ctx)
            else:
                # create one task per call
                for item in rows_call_args:
                    task = asyncio.create_task(self.eval_async(item))
                    self.dispatcher.register_task(task)

        else:
            # create a single task for all rows
            task = asyncio.create_task(self.eval(rows_call_args))
            self.dispatcher.register_task(task)

    def _queued_call_args_iter(self) -> Iterator[FnCallArgs]:
        while not self.call_args_queue.empty():
            yield self.call_args_queue.get_nowait()

    def _create_batch_call_args(self, call_args: list[FnCallArgs]) -> FnCallArgs:
        """Roll call_args into a single batched FnCallArgs"""
        batch_args: list[list[Optional[Any]]] = [[None] * len(call_args) for _ in range(len(self.fn_call.arg_idxs))]
        batch_kwargs: dict[str, list[Optional[Any]]] = {k: [None] * len(call_args) for k in self.fn_call.kwarg_idxs}
        assert isinstance(self.fn, func.CallableFunction)
        for i, item in enumerate(call_args):
            for j in range(len(item.args)):
                batch_args[j][i] = item.args[j]
            for k in item.kwargs:
                batch_kwargs[k][i] = item.kwargs[k]
        return FnCallArgs(
            self.fn_call, [item.row for item in call_args], batch_args=batch_args, batch_kwargs=batch_kwargs
        )

    async def eval_batch(self, batched_call_args: FnCallArgs) -> None:
        result_batch: list[Any]
        try:
            if self.fn.is_async:
                result_batch = await self.fn.aexec_batch(
                    *batched_call_args.batch_args, **batched_call_args.batch_kwargs
                )
            else:
                # check for cancellation before starting something potentially long-running
                if asyncio.current_task().cancelled() or self.dispatcher.exc_event.is_set():
                    return
                result_batch = self.fn.exec_batch(batched_call_args.batch_args, batched_call_args.batch_kwargs)
        except Exception as exc:
            _, _, exc_tb = sys.exc_info()
            for row in batched_call_args.rows:
                row.set_exc(self.fn_call.slot_idx, exc)
            self.dispatcher.dispatch_exc(batched_call_args.rows, self.fn_call.slot_idx, exc_tb, self.exec_ctx)
            return

        for i, row in enumerate(batched_call_args.rows):
            row[self.fn_call.slot_idx] = result_batch[i]
        self.dispatcher.dispatch(batched_call_args.rows, self.exec_ctx)

    async def eval_async(self, call_args: FnCallArgs) -> None:
        assert len(call_args.rows) == 1
        assert not call_args.row.has_val[self.fn_call.slot_idx]
        assert not call_args.row.has_exc(self.fn_call.slot_idx)

        try:
            start_ts = datetime.datetime.now()
            _logger.debug(f'Start evaluating slot {self.fn_call.slot_idx}')
            call_args.row[self.fn_call.slot_idx] = await self.fn.aexec(*call_args.args, **call_args.kwargs)
            end_ts = datetime.datetime.now()
            _logger.debug(f'Evaluated slot {self.fn_call.slot_idx} in {end_ts - start_ts}')
            self.dispatcher.dispatch([call_args.row], self.exec_ctx)
        except Exception as exc:
            import anthropic

            if isinstance(exc, anthropic.RateLimitError):
                _logger.debug(f'RateLimitError: {exc}')
            _, _, exc_tb = sys.exc_info()
            call_args.row.set_exc(self.fn_call.slot_idx, exc)
            self.dispatcher.dispatch_exc(call_args.rows, self.fn_call.slot_idx, exc_tb, self.exec_ctx)

    async def eval(self, call_args_batch: list[FnCallArgs]) -> None:
        rows_with_excs: set[int] = set()  # records idxs into 'rows'
        for idx, item in enumerate(call_args_batch):
            assert len(item.rows) == 1
            assert not item.row.has_val[self.fn_call.slot_idx]
            assert not item.row.has_exc(self.fn_call.slot_idx)
            # check for cancellation before starting something potentially long-running
            if asyncio.current_task().cancelled() or self.dispatcher.exc_event.is_set():
                return
            try:
                item.row[self.fn_call.slot_idx] = self.scalar_py_fn(*item.args, **item.kwargs)
            except Exception as exc:
                _, _, exc_tb = sys.exc_info()
                item.row.set_exc(self.fn_call.slot_idx, exc)
                rows_with_excs.add(idx)
                self.dispatcher.dispatch_exc(item.rows, self.fn_call.slot_idx, exc_tb, self.exec_ctx)
        self.dispatcher.dispatch(
            [call_args_batch[i].row for i in range(len(call_args_batch)) if i not in rows_with_excs], self.exec_ctx
        )

    def _close(self) -> None:
        """Create a task for the incomplete batch of queued FnCallArgs, if any"""
        _logger.debug(f'FnCallEvaluator.close(): slot_idx={self.fn_call.slot_idx}')
        if self.call_args_queue is None or self.call_args_queue.empty():
            return
        batched_call_args = self._create_batch_call_args(list(self._queued_call_args_iter()))
        task = asyncio.create_task(self.eval_batch(batched_call_args))
        self.dispatcher.register_task(task)


class NestedRowList:
    """
    A list of nested rows, used by JsonMapperDispatcher to store the rows corresponding to the elements of the
    JsonMapper source list and make completion awaitable.
    """

    rows: list[exprs.DataRow]
    num_completed: int
    completion: asyncio.Event

    def __init__(self, rows: list[exprs.DataRow]):
        self.num_completed = 0
        self.rows = rows
        self.completion = asyncio.Event()

    def complete_row(self) -> None:
        self.num_completed += 1
        if self.num_completed == len(self.rows):
            self.completion.set()


class JsonMapperDispatcher(Evaluator):
    """
    The execution logic for materializing the nested DataRows of a JsonMapper/JsonMapperDispatch.

    The rows are stored in a NestedRowList, which itself is stored in the JsonMapperDispatch instance's slot.
    """

    e: exprs.JsonMapperDispatch
    target_expr: exprs.Expr
    scope_anchor: exprs.ObjectRef
    nested_exec_ctx: ExecCtx  # ExecCtx needed to evaluate the nested rows
    external_slot_map: dict[int, int]  # slot idx in parent row -> slot idx in nested row
    has_async_calls: bool  # True if target_expr contains any async FunctionCalls

    def __init__(self, e: exprs.JsonMapperDispatch, dispatcher: Dispatcher, exec_ctx: ExecCtx):
        super().__init__(dispatcher, exec_ctx)
        self.e = e
        self.target_expr = e.target_expr.copy()  # we need new slot idxs
        self.scope_anchor = e.scope_anchor.copy()
        nested_row_builder = exprs.RowBuilder(output_exprs=[self.target_expr], columns=[], input_exprs=[])
        nested_row_builder.set_slot_idxs([self.target_expr, self.scope_anchor])
        target_expr_ctx = nested_row_builder.create_eval_ctx([self.target_expr], limit_scope=True)
        self.has_async_calls = any(isinstance(e, exprs.FunctionCall) and e.is_async for e in target_expr_ctx.exprs)
        target_scope = self.target_expr.scope()
        # we need to pre-populated nested rows with slot values that are produced in an outer scope (literals excluded)
        parent_exprs = [
            e for e in target_expr_ctx.exprs if e.scope() != target_scope and not isinstance(e, exprs.Literal)
        ]
        self.external_slot_map = {exec_ctx.row_builder.unique_exprs[e].slot_idx: e.slot_idx for e in parent_exprs}
        self.nested_exec_ctx = ExecCtx(dispatcher, nested_row_builder, [self.target_expr], parent_exprs)

    def schedule(self, rows: list[exprs.DataRow], slot_idx: int) -> None:
        """Create nested rows for all source list elements and dispatch them"""
        assert self.e.slot_idx >= 0
        all_nested_rows: list[exprs.DataRow] = []
        for row in rows:
            src = row[self.e.src_expr.slot_idx]
            if not isinstance(src, list):
                # invalid/non-list src path
                row[self.e.slot_idx] = None
                continue

            nested_rows = [
                exprs.DataRow(
                    size=self.nested_exec_ctx.row_builder.num_materialized,
                    img_slot_idxs=[],
                    media_slot_idxs=[],
                    array_slot_idxs=[],
                    parent_row=row,
                    parent_slot_idx=self.e.slot_idx,
                )
                for _ in src
            ]
            for nested_row, anchor_val in zip(nested_rows, src):
                nested_row[self.scope_anchor.slot_idx] = anchor_val
                for slot_idx_, nested_slot_idx in self.external_slot_map.items():
                    nested_row[nested_slot_idx] = row[slot_idx_]
            self.nested_exec_ctx.init_rows(nested_rows)

            # we modify DataRow.vals here directly, rather than going through __getitem__(), because we don't have
            # an official "value" yet (the nested rows are not yet materialized)
            row.vals[self.e.slot_idx] = NestedRowList(nested_rows)
            all_nested_rows.extend(nested_rows)

        self.dispatcher.dispatch(all_nested_rows, self.nested_exec_ctx)
        task = asyncio.create_task(self.gather(rows))
        self.dispatcher.register_task(task)

    async def gather(self, rows: list[exprs.DataRow]) -> None:
        """Wait for nested rows to complete, then signal completion to the parent rows"""
        if self.has_async_calls:
            # if our target expr contains async FunctionCalls, they typically get completed out-of-order, and it's
            # more effective to dispatch them as they complete
            remaining = {
                asyncio.create_task(row.vals[self.e.slot_idx].completion.wait()): row
                for row in rows
                if not row.has_val[self.e.slot_idx]
            }
            while len(remaining) > 0:
                done, _ = await asyncio.wait(remaining.keys(), return_when=asyncio.FIRST_COMPLETED)
                done_rows = [remaining.pop(task) for task in done]
                for row in done_rows:
                    row.has_val[self.e.slot_idx] = True
                self.dispatcher.dispatch(done_rows, self.exec_ctx)

        else:
            # our target expr doesn't contain async FunctionCalls, which means they will get completed in-order
            for row in rows:
                if row.has_val[self.e.slot_idx]:
                    # the source_expr's value is not a list
                    assert row.vals[self.e.slot_idx] is None
                    continue
                assert row.vals[self.e.slot_idx] is not None and isinstance(row.vals[self.e.slot_idx], NestedRowList)
                await row.vals[self.e.slot_idx].completion.wait()
                row.has_val[self.e.slot_idx] = True
            self.dispatcher.dispatch(rows, self.exec_ctx)



"""
Extended integrations for Pixeltable. This package contains experimental or demonstration features that
are not intended for production use. Long-term support cannot be guaranteed, usually because the features
have dependencies whose future support is unclear.
"""

# ruff: noqa: F401

from pixeltable.utils.code import local_public_names

from . import functions

__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import os
from typing import Any, Iterator, Optional, Union

import fiftyone as fo  # type: ignore[import-untyped]
import fiftyone.utils.data as foud  # type: ignore[import-untyped]
import PIL.Image
import puremagic

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable import exprs
from pixeltable.env import Env


class PxtImageDatasetImporter(foud.LabeledImageDatasetImporter):
    """
    Implementation of a FiftyOne `DatasetImporter` that reads image data from a Pixeltable table.
    """

    __image_format: str  # format to use for any exported images that are not already stored on disk
    __labels: dict[str, tuple[exprs.Expr, type[fo.Label]]]  # label_name -> (expr, label_cls)
    __image_idx: int  # index of the image expr in the select list
    __localpath_idx: Optional[int]  # index of the image localpath in the select list, if present
    __row_iter: Iterator[list]  # iterator over the table rows, to be convered to FiftyOne samples

    def __init__(
        self,
        tbl: pxt.Table,
        image: exprs.Expr,
        image_format: str,
        classifications: Union[exprs.Expr, list[exprs.Expr], dict[str, exprs.Expr], None] = None,
        detections: Union[exprs.Expr, list[exprs.Expr], dict[str, exprs.Expr], None] = None,
        dataset_dir: Optional[os.PathLike] = None,
        shuffle: bool = False,
        seed: Union[int, float, str, bytes, bytearray, None] = None,
        max_samples: Optional[int] = None,
    ):
        super().__init__(dataset_dir=dataset_dir, shuffle=shuffle, seed=seed, max_samples=max_samples)

        self.__image_format = image_format

        label_categories = [
            (classifications, fo.Classifications, 'classifications'),
            (detections, fo.Detections, 'detections'),
        ]

        # Construct the labels. First add labels for all label types that have named dictionaries.
        self.__labels = {}
        for exprs_, label_cls, _ in label_categories:
            if isinstance(exprs_, dict):
                for label_name, expr in exprs_.items():
                    if not label_name.isidentifier():
                        raise excs.Error(f'Invalid label name: {label_name}')
                    if label_name in self.__labels:
                        raise excs.Error(f'Duplicate label name: {label_name}')
                    self.__labels[label_name] = (expr, label_cls)

        # Now add the remaining labels, assigning unused default names.
        for exprs_, label_cls, default_name in label_categories:
            if exprs_ is None or isinstance(exprs_, dict):
                continue
            exprs_list = [exprs_] if isinstance(exprs_, exprs.Expr) else exprs_
            assert isinstance(exprs_list, list)
            for expr in exprs_list:
                if default_name not in self.__labels:
                    name = default_name
                else:
                    i = 1
                    while f'{default_name}_{i}' in self.__labels:
                        i += 1
                    name = f'{default_name}_{i}'
                self.__labels[name] = (expr, label_cls)

        # Build the select list:
        # - Labels first, in the order they appear in self.__labels
        # - Then the `image` expr
        # - Then `image.localpath`, if `images` is a stored columnref

        selection = [expr for expr, _ in self.__labels.values()]
        self.__image_idx = len(selection)
        selection.append(image)

        if isinstance(image, exprs.ColumnRef) and image.col.is_stored:
            # A stored image column; we can use the existing localpaths
            self.__localpath_idx = len(selection)
            selection.append(image.localpath)
        else:
            self.__localpath_idx = None

        df = tbl.select(*selection)
        self.__row_iter = df._output_row_iterator()

    def __next__(self) -> tuple[str, Optional[fo.ImageMetadata], Optional[dict[str, fo.Label]]]:
        row = next(self.__row_iter)
        img = row[self.__image_idx]
        assert isinstance(img, PIL.Image.Image)
        if self.__localpath_idx is not None:
            # Use the existing localpath of the stored image
            file = row[self.__localpath_idx]
            assert isinstance(file, str)
        else:
            # Write the dynamically created image to a temp file
            file = str(Env.get().create_tmp_path(f'.{self.__image_format}'))
            img.save(file, format=self.__image_format)

        metadata = fo.ImageMetadata(
            size_bytes=os.path.getsize(file),
            mime_type=puremagic.from_file(file, mime=True),
            width=img.width,
            height=img.height,
            filepath=file,
            num_channels=len(img.getbands()),
        )

        labels: dict[str, fo.Label] = {}
        for idx, (label_name, (_, label_cls)) in enumerate(self.__labels.items()):
            label_data = row[idx]
            if label_data is None:
                continue

            label: fo.Label
            if label_cls is fo.Classifications:
                label = fo.Classifications(classifications=self.__as_fo_classifications(label_data))
            elif label_cls is fo.Detections:
                label = fo.Detections(detections=self.__as_fo_detections(label_data))
            else:
                raise AssertionError()
            labels[label_name] = label

        return file, metadata, labels

    def __as_fo_classifications(self, data: list) -> list[fo.Classification]:
        if not isinstance(data, list) or any('label' not in entry for entry in data):
            raise excs.Error(
                f"Invalid classifications data: {data}\n(Expected a list of dicts, each containing a 'label' key)"
            )
        return [fo.Classification(label=entry['label'], confidence=entry.get('confidence')) for entry in data]

    def __as_fo_detections(self, data: list) -> list[fo.Detections]:
        if not isinstance(data, list) or any('label' not in entry or 'bounding_box' not in entry for entry in data):
            raise excs.Error(
                f'Invalid detections data: {data}\n'
                "(Expected a list of dicts, each containing a 'label' and 'bounding_box' key)"
            )
        return [
            fo.Detection(label=entry['label'], bounding_box=entry['bounding_box'], confidence=entry.get('confidence'))
            for entry in data
        ]

    @property
    def has_dataset_info(self) -> bool:
        return False

    @property
    def has_image_metadata(self) -> bool:
        return True

    @property
    def label_cls(self) -> dict[str, type]:
        return {label_name: label_cls for label_name, (_, label_cls) in self.__labels.items()}

    def setup(self) -> None:
        pass

    def get_dataset_info(self) -> dict:
        pass

    def close(self, *args: Any) -> None:
        pass



import json
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import httpx

import pixeltable as pxt
from pixeltable import env
from pixeltable.utils.code import local_public_names

from .openai import _opt

if TYPE_CHECKING:
    import openai


@env.register_client('deepseek')
def _(api_key: str) -> 'openai.AsyncOpenAI':
    import openai

    return openai.AsyncOpenAI(
        api_key=api_key,
        base_url='https://api.deepseek.com',
        http_client=httpx.AsyncClient(limits=httpx.Limits(max_keepalive_connections=100, max_connections=500)),
    )


def _deepseek_client() -> 'openai.AsyncOpenAI':
    return env.Env.get().get_client('deepseek')


@pxt.udf
async def chat_completions(
    messages: list,
    *,
    model: str,
    frequency_penalty: Optional[float] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    max_tokens: Optional[int] = None,
    presence_penalty: Optional[float] = None,
    response_format: Optional[dict] = None,
    stop: Optional[list[str]] = None,
    temperature: Optional[float] = None,
    tools: Optional[list[dict]] = None,
    tool_choice: Optional[dict] = None,
    top_p: Optional[float] = None,
) -> dict:
    """
    Creates a model response for the given chat conversation.

    Equivalent to the Deepseek `chat/completions` API endpoint.
    For additional details, see: <https://api-docs.deepseek.com/api/create-chat-completion>

    Deepseek uses the OpenAI SDK, so you will need to install the `openai` package to use this UDF.

    __Requirements:__

    - `pip install openai`

    Args:
        messages: A list of messages to use for chat completion, as described in the Deepseek API documentation.
        model: The model to use for chat completion.

    For details on the other parameters, see: <https://api-docs.deepseek.com/api/create-chat-completion>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `deepseek-chat` to an existing Pixeltable column `tbl.prompt`
        of the table `tbl`:

        >>> messages = [
                {'role': 'system', 'content': 'You are a helpful assistant.'},
                {'role': 'user', 'content': tbl.prompt}
            ]
            tbl.add_computed_column(response=chat_completions(messages, model='deepseek-chat'))
    """
    if tools is not None:
        tools = [{'type': 'function', 'function': tool} for tool in tools]

    tool_choice_: Union[str, dict, None] = None
    if tool_choice is not None:
        if tool_choice['auto']:
            tool_choice_ = 'auto'
        elif tool_choice['required']:
            tool_choice_ = 'required'
        else:
            assert tool_choice['tool'] is not None
            tool_choice_ = {'type': 'function', 'function': {'name': tool_choice['tool']}}

    extra_body: Optional[dict[str, Any]] = None
    if tool_choice is not None and not tool_choice['parallel_tool_calls']:
        extra_body = {'parallel_tool_calls': False}

    # cast(Any, ...): avoid mypy errors
    result = await _deepseek_client().chat.completions.with_raw_response.create(
        messages=messages,
        model=model,
        frequency_penalty=_opt(frequency_penalty),
        logprobs=_opt(logprobs),
        top_logprobs=_opt(top_logprobs),
        max_tokens=_opt(max_tokens),
        presence_penalty=_opt(presence_penalty),
        response_format=_opt(cast(Any, response_format)),
        stop=_opt(stop),
        temperature=_opt(temperature),
        tools=_opt(cast(Any, tools)),
        tool_choice=_opt(cast(Any, tool_choice_)),
        top_p=_opt(top_p),
        extra_body=extra_body,
    )

    return json.loads(result.text)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__
